//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_Dapi_register_commands.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
// 
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	12.11.2013	Start DEDITEC-History

#ifdef USE_ETH_INTERN
// Muss sein, wegen return als long long
extern unsigned long long DapiROETHInternReadLongLong(unsigned long handle32, unsigned long adress);
#endif

// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiPing
// ------------------------------------
// DapiWriteByte
// DapiWriteWord
// DapiWriteLong
// DapiWriteLongLong
// ------------------------------------
// DapiReadByte
// DapiReadWord
// DapiReadLong
// DapiReadLongLong
// ------------------------------------
// DapiReadMultipleBytes
// DapiWriteMultipleBytes
// ------------------------------------
// DapiReadFifo

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	#ifdef USE_FTDI
		#ifdef USE_USB_SIMPLE
			#include "../usb-simple/ftd2xx.h"
		#endif
		#ifdef USE_USB_FTDI 
			#include "../usb-ftdi/ftd2xx.h"
		#endif
	#endif

	#include "library/delib/delib.h"
	#include "library/vc/deditec_tcp_io.h"	// muss vor die "delib_intern.h", weil hier TCP_IO drin ist
	#include "library/delib/delib_intern.h"
	
	#include "delib_io_seriellFTDI.h"
	#include "delib_io_ftdi.h"
	#include "delib_io_tcp.h"
	#include "delib_set_get_error.h"
	#include "delib_hdlmgr.h"
	#include "delib_open_close.h"
	
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include <windows.h>
	#include "conio.h"
	#include "malloc.h"
	
	#include "delib_io_seriellCOM.h"
	#include "delib_io_seriellFTDI.h"
	#include "delib_io_ftdi.h"
	#include "delib_io_tcp.h"
	#include "delib_io_CANopen.h"
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
#endif

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiPing(ULONG handle32, ULONG value)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	char dapi_caller_text[200];

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiPing(handle=0x%lx, value=0x%lx)", (unsigned long) handle32, (unsigned long) value);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		
		DAPIx_return(0);
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiPing");
		sprintf(debug_msg,"DapiPing(handle32=0x%lx, value=0x%lx)", (unsigned long) handle32, (unsigned long) value);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				DapiSeriellFTDIWriteByte(handle32, 0xffff, value);		// Save value at Adress 0xffff
				ret=DapiSeriellFTDIReadByte(handle32, 0xffff);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				ret = DapiFTDIPing(handle32, value);
				break;
			// ------------------------------------
		#endif

		#ifdef USE_SER
		// ------------------------------------
			case bus_SeriellCOM:
				DapiSeriellCOMWriteByte(handle32, 0xffff, value);			// Save value at Adress 0xffff
				ret=DapiSeriellCOMReadByte(handle32, 0xffff);
				break;
		#endif

		#ifdef USE_ETH
		// ------------------------------------
			case bus_ETH:
				DapiTCPWriteByte(handle32, 0xffff, value);				// Save value at Adress 0xffff
				ret=DapiTCPReadByte(handle32, 0xffff);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
		// ------------------------------------
			case bus_ETH_INTERN:
				DapiROETHInternWriteByte(handle32, 0xffff, value);		// Save value at Adress 0xffff
				ret=DapiROETHInternReadByte(handle32, 0xffff);
				break;
		#endif
		
		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}




// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteByte(ULONG handle, ULONG adress, ULONG value)
{
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiWriteByte(handle=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle, (unsigned long) adress, (unsigned long) value);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	DapiWriteByteXXX(handle, adress, value);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

}

// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteWord(ULONG handle, ULONG adress, ULONG value)
{
	char dapi_caller_text[200];

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiWriteWord(handle=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle, (unsigned long) adress, (unsigned long) value);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	DapiWriteWordXXX(handle, adress, value);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteLong(ULONG handle, ULONG adress, ULONG value)
{
	char dapi_caller_text[200];

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiWriteLong(handle=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle, (unsigned long) adress, (unsigned long) value);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	DapiWriteLongXXX(handle, adress, value);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteLongLong(ULONG handle, ULONG adress, ULONGLONG value)
{
	char dapi_caller_text[200];

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiWriteLongLong(handle=0x%lx, address=0x%lx, value=0x%llx)", (unsigned long) handle, (unsigned long) adress, (unsigned long long) value);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	DapiWriteLongLongXXX( handle, adress, value);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

}

// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadByte(ULONG handle, ULONG adress)
{
	char dapi_caller_text[200];
	unsigned long ret;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiReadByte(handle=0x%lx, address=0x%lx)", (unsigned long) handle, (unsigned long) adress);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiReadByteXXX(handle, adress);
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadWord(ULONG handle, ULONG adress)
{
	char dapi_caller_text[200];
	unsigned long ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiReadWord(handle=0x%lx, address=0x%lx)", (unsigned long) handle, (unsigned long) adress);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiReadWordXXX(handle, adress);
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadLong(ULONG handle, ULONG adress)
{
	char dapi_caller_text[200];
	unsigned long ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiReadLong(handle=0x%lx, address=0x%lx)", (unsigned long) handle, (unsigned long) adress);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiReadLongXXX(handle, adress);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------

ULONGLONG DAPI_FUNCTION_PRE DapiReadLongLong(ULONG handle, ULONG adress)
{
	char dapi_caller_text[200];
	unsigned long long ret;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiReadLongLong(handle=0x%lx, address=0x%lx)", (unsigned long) handle, (unsigned long) adress);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiReadLongLongXXX(handle, adress);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadMultipleBytes(ULONG handle, ULONG adress, ULONG adress_depth, ULONG repeat, UCHAR * buff, ULONG buff_len)
{
	char dapi_caller_text[200];
	unsigned long ret;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiReadMultipleBytes(handle=0x%lx, address=0x%lx, address_depth=0x%lx, repeat=0x%lx, buff_len=0x%lx)", (unsigned long) handle, (unsigned long) adress, (unsigned long) adress_depth, (unsigned long) repeat, (unsigned long) buff_len);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiReadMultipleBytesXXX(handle, adress, adress_depth, repeat, buff, buff_len);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiWriteMultipleBytes(ULONG handle, ULONG adress, ULONG adress_depth, ULONG repeat, UCHAR * buff, ULONG buff_len)
{
	char dapi_caller_text[200];
	unsigned long ret;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiWriteMultipleBytes(handle=0x%lx, address=0x%lx, address_depth=0x%lx, repeat=0x%lx, buff_len=0x%lx)", (unsigned long) handle, (unsigned long) adress, (unsigned long) adress_depth, (unsigned long) repeat, (unsigned long) buff_len);
	DapiSetErrorCallerParamsSet(handle, dapi_caller_text, sizeof(dapi_caller_text));

	ret=DapiWriteMultipleBytesXXX(handle, adress, adress_depth, repeat, buff, buff_len);
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteByteXXX(ULONG handle32, ULONG adress, ULONG value)
{
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiWriteByte");
		sprintf(debug_msg,"DapiWriteByte(handle=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle32, (unsigned long) adress, (unsigned long) value);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				DapiSeriellFTDIWriteByte(handle32, adress, value);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					DapiFTDIWriteByte(handle32, adress, value);
				}
				else
				{
					#ifndef __linux__
					DapiCANopenWriteByte(handle32, adress, value);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				DapiSeriellCOMWriteByte(handle32, adress, value);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				DapiTCPWriteByte(handle32, adress, value);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				DapiROETHInternWriteByte(handle32, adress, value);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteWordXXX(ULONG handle32, ULONG adress, ULONG value)
{
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiWriteWord");
		sprintf(debug_msg,"DapiWriteWord(handle=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle32, (unsigned long) adress, (unsigned long) value);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				DapiSeriellFTDIWriteWord(handle32, adress, value);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					DapiFTDIWriteWord(handle32, adress, value);
				}
				else
				{
					#ifndef __linux__
					DapiCANopenWriteWord(handle32, adress, value);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				DapiSeriellCOMWriteWord(handle32, adress, value);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				DapiTCPWriteWord(handle32, adress, value);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				DapiROETHInternWriteWord(handle32, adress, value);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteLongXXX(ULONG handle32, ULONG adress, ULONG value)
{
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiWriteLong");
		sprintf(debug_msg,"DapiWriteLong(handle32=0x%lx, address=0x%lx, value=0x%lx)", (unsigned long) handle32, (unsigned long) adress, (unsigned long) value);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				DapiSeriellFTDIWriteLong(handle32, adress, value);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					DapiFTDIWriteLong(handle32, adress, value);
				}
				else
				{
					#ifndef __linux__
					DapiCANopenWriteLong(handle32, adress, value);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				DapiSeriellCOMWriteLong(handle32, adress, value);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				DapiTCPWriteLong(handle32, adress, value);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				DapiROETHInternWriteLong(handle32, adress, value);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DAPI_FUNCTION_PRE DapiWriteLongLongXXX(ULONG handle32, ULONG adress, ULONGLONG value)
{
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiWriteLongLong");
		#ifdef __linux__
			sprintf(debug_msg,"DapiWriteLongLong(handle32=0x%lx, address=0x%lx, value=0x%llx)", (unsigned long) handle32, (unsigned long) adress, value);
		#else
			sprintf(debug_msg,"DapiWriteLongLong(handle32=0x%lx, address=0x%lx, value=0x%I64x)", (unsigned long) handle32, (unsigned long) adress, value);
		#endif
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				DapiSeriellFTDIWriteLongLong(handle32, adress, value);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					DapiFTDIWriteLongLong(handle32, adress, value);
				}
				else
				{
					#ifndef __linux__
					DapiCANopenWriteLongLong(handle32, adress, value);
					#endif
				}
				break;
		#endif


		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				DapiSeriellCOMWriteLongLong(handle32, adress, value);
				break;
		#endif


		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				DapiTCPWriteLongLong(handle32, adress, value);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				DapiROETHInternWriteLongLong(handle32, adress, value);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadByteXXX(ULONG handle32, ULONG adress)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiReadByte");
		sprintf(debug_msg,"DapiReadByte(handle32=0x%lx, address=0x%lx)", (unsigned long) handle32, (unsigned long) adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				ret = DapiSeriellFTDIReadByte(handle32, adress);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					ret = DapiFTDIReadByte(handle32, adress);
				}
				else
				{
					#ifndef __linux__
					ret = DapiCANopenReadByte(handle32, adress);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				ret = DapiSeriellCOMReadByte(handle32, adress);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				ret = DapiTCPReadByte(handle32, adress);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternReadByte(handle32, adress);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadByte");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadWordXXX(ULONG handle32, ULONG adress)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiReadWord");
		sprintf(debug_msg,"DapiReadWord(handle32=0x%lx, address=0x%lx)", (unsigned long) handle32, (unsigned long) adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				ret = DapiSeriellFTDIReadWord(handle32, adress);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					ret = DapiFTDIReadWord(handle32, adress);
				}
				else
				{
					#ifndef __linux__
					ret = DapiCANopenReadWord(handle32, adress);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				ret = DapiSeriellCOMReadWord(handle32, adress);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				ret = DapiTCPReadWord(handle32, adress);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternReadWord(handle32, adress);
				break;
		#endif
		
		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadWord");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	return ret;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadLongXXX(ULONG handle32, ULONG adress)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiReadLong");
		sprintf(debug_msg,"DapiReadLong(handle32=0x%lx, address=0x%lx)", (unsigned long) handle32, (unsigned long) adress);
		debug_print(debug_msg);
	#endif
	// -----

	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
		// ------------------------------------
			case bus_SeriellFTDI:
				ret = DapiSeriellFTDIReadLong(handle32, adress);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					ret = DapiFTDIReadLong(handle32, adress);
				}
				else
				{
					#ifndef __linux__
					ret = DapiCANopenReadLong(handle32, adress);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				ret = DapiSeriellCOMReadLong(handle32, adress);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				ret = DapiTCPReadLong(handle32, adress);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternReadLong(handle32, adress);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadLong");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	return ret;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONGLONG DAPI_FUNCTION_PRE DapiReadLongLongXXX(ULONG handle32, ULONG adress)
{
	ULONGLONG ret;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiReadLongLong");
		sprintf(debug_msg,"DapiReadLongLong(handle32=0x%lx, address=0x%lx)", (unsigned long) handle32, (unsigned long) adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				ret = DapiSeriellFTDIReadLongLong(handle32, adress);
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				if(((adress&0xffff0000) != 0) || (DapiHandle->CANopenStickSondermodus == 0))
				{
					ret = DapiFTDIReadLongLong(handle32, adress);
				}
				else
				{
					#ifndef __linux__
					ret = DapiCANopenReadLongLong(handle32, adress);
					#endif
				}
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				ret = DapiSeriellCOMReadLongLong(handle32, adress);
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				ret = DapiTCPReadLongLong(handle32, adress);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternReadLongLong(handle32, adress);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadLongLong");
		#ifdef __linux__
			sprintf(debug_msg,"return=0x%llx", ret);
		#else
			sprintf(debug_msg,"return=0x%I64x", ret);
		#endif
		debug_print(debug_msg);
	#endif
	// -----

	return ret;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiReadMultipleBytesXXX(ULONG handle32, ULONG adress, ULONG adress_depth, ULONG repeat, UCHAR *buff, ULONG buff_size)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadMultipleBytes");
		sprintf(debug_msg,"DapiReadMultipleBytes(handle32=0x%lx, address=0x%lx, address_depth=0x%lx, repeat=0x%lx, buff_size=0x%lx)", (unsigned long) handle32, (unsigned long) adress, (unsigned long) adress_depth, (unsigned long) repeat, (unsigned long) buff_size);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
		// ------------------------------------
			case bus_SeriellFTDI:
				#ifndef __linux__
					if (buff_size > 64)
					{
						sprintf(debug_msg, "DapiSeriellFTDIRead64BytesMax: buffer too large * buff_len = %ld [dec]", (unsigned long) buff_size);
						DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
						return RETURN_ERROR;
					}
				
					DapiSeriellFTDIRead64BytesMax(handle32, adress, buff, adress_depth);
				#endif
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				ret = DapiFTDIReadMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_size);
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:			
				if (buff_size > 64)
				{
					sprintf(debug_msg, "DapiSeriellCOMRead64BytesMax: buffer too large * buff_len = %ld [dec]", (unsigned long) buff_size);
					DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
					return RETURN_ERROR;
				}
				
				DapiSeriellCOMRead64BytesMax(handle32, adress, buff, adress_depth);
				break;
		#endif
		
		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				// only ro-eth-modules with software-features support this command!
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0)
				{
					sprintf(debug_msg, "DapiReadMultipleBytes -> Module does not support this command    DapiModuleParams->sw_feature1=0x%lx", (unsigned long) DapiModuleParams->sw_feature1);
					DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
					return RETURN_ERROR;
				}
			
				ret = DapiTCPReadMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_size);
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternReadMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_len);
				break;
		#endif
		
		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadMultipleBytes");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DAPI_FUNCTION_PRE DapiWriteMultipleBytesXXX(ULONG handle32, ULONG adress, ULONG adress_depth, ULONG repeat, UCHAR * buff, ULONG buff_len)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG len;
	ULONG pos;
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiWriteMultipleBytes");
		sprintf(debug_msg,"DapiWriteMultipleBytes(handle32=0x%lx, address=0x%lx, address_depth=0x%lx, repeat=0x%lx, buff_len=0x%lx)", (unsigned long) handle32, (unsigned long) adress, (unsigned long) adress_depth, (unsigned long) repeat, (unsigned long) buff_len);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_FTDI
			// ------------------------------------
			case bus_SeriellFTDI:
				if (buff_len > 1024)
				{
					sprintf(debug_msg, "DapiSeriellFTDIWrite64BytesMax: buffer too large ( > 1024 Bytes) * buff_len = %ld [dec]", (unsigned long) buff_len);
					DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
					return RETURN_ERROR;
				}
				
				len = buff_len;
				pos = 0;
				while(len > 0)
				{
					if(len > 32)// Mit 32 klappts besser !
					{
						DapiSeriellFTDIWrite64BytesMax(handle32, adress + pos, buff + pos, 32);
						pos += 32;
						len -= 32;
					}
					else
					{
						DapiSeriellFTDIWrite64BytesMax(handle32, adress + pos, buff + pos, len);
						pos += len;
						len = 0;
					}
				}
				break;
			// ------------------------------------
			case bus_FTDI_DUAL:
			case bus_FTDI_SINGLE_SER:
			case bus_FTDI_SINGLE_PAR:
				ret = DapiFTDIWriteMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_len);
				break;
		#endif

		#ifdef USE_SER
			// ------------------------------------
			case bus_SeriellCOM:
				if (buff_len > 1024)
				{
					sprintf(debug_msg, "DapiSeriellCOMWrite64BytesMax: buffer too large ( > 1024 Bytes) * buff_len = %ld [dec]", (unsigned long) buff_len);
					DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
					return RETURN_ERROR;
				}
			
				len = buff_len;
				pos = 0;
				while(len > 0)
				{
					if(len > 64)
					{
						DapiSeriellCOMWrite64BytesMax(handle32, adress + pos, buff + pos, 64);
						pos += 64;
						len -= 64;
					}
					else
					{
						DapiSeriellCOMWrite64BytesMax(handle32, adress + pos, buff + pos, len);
						pos += len;
						len = 0;
					}
				}
				break;
		#endif

		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				// only ro-eth-modules with software-features support this command!
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0)
				{
					//sprintf(debug_msg, "DapiWriteMultipleBytes -> Module does not support this command");
					//DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
					//return RETURN_ERROR;
					ret = DapiTCPWriteMultipleBytesOLD(handle32, adress, adress_depth, repeat, buff, buff_len);
				}
				else
				{
					ret = DapiTCPWriteMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_len);
				}
				break;
		#endif
		
		#ifdef USE_ETH_INTERN
			// ------------------------------------
			case bus_ETH_INTERN:
				ret = DapiROETHInternWriteMultipleBytes(handle32, adress, adress_depth, repeat, buff, buff_len);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiWriteMultipleBytes");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiReadFifo(ULONG handle32, ULONG type, UCHAR *buff, ULONG buff_len)
{
	ULONG ret;
	DAPI_HANDLE *DapiHandle;
	char dapi_caller_text[200];

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiReadFifo(handle=0x%lx, type=0x%lx, buff_len=0x%lx)", (unsigned long) handle32, (unsigned long) type, (unsigned long) buff_len);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiReadFifo");
		sprintf(debug_msg,"DapiReadFifo(handle32=0x%lx, type=0x%lx, buff_len=0x%lx)", (unsigned long) handle32, (unsigned long) type, (unsigned long) buff_len);
		debug_print(debug_msg);
	#endif
	// -----
	
	switch(DapiHandle->BusType)
	{
		#ifdef USE_ETH
			// ------------------------------------
			case bus_ETH:
				ret = DapiTCPReadFifo(handle32, type, buff, buff_len);
				break;
		#endif

		// ------------------------------------
		default:
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, NULL, __FILE__, __LINE__);					
			break;
		// ------------------------------------
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiReadFifo");
		sprintf(debug_msg,"return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

