//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_io_ftdi.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiOpenModuleIntern_FTDI
// IsFTDIError
// DapiFTDIdoReConnect
// ------------------------------------
// DapiFTDISetModuleAdress
// DapiOpenModuleGetFT_HANDLE
// FTread_n_bytes_with_timeout
// DapiFTDISendCommand
// DapiFTDISendCommandMultipleBytes
// ------------------------------------
// DapiFTDIPing
// ------------------------------------
// DapiFTDIWriteByte
// DapiFTDIWriteWord
// DapiFTDIWriteLong
// DapiFTDIWriteLongLong
// ------------------------------------
// DapiFTDIReadByte
// DapiFTDIReadWord
// DapiFTDIReadLong
// DapiFTDIReadLongLong
// ------------------------------------
// DapiFTDIReadMultipleBytes
// DapiFTDIWriteMultipleBytes

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	#ifdef USE_USB_SIMPLE
		#include "../usb-simple/ftdi.h"
		#include "../usb-simple/ftd2xx.h"
		#include "../usb-simple/ftdi_functions.c"
	#endif
	
	#ifdef USE_USB_FTDI
		#include "../usb-ftdi/ftd2xx.h"
	#endif
	
	#include "library/vc/deditec_tcp_io.h"
	#include "library/vc/deditec_convert_functions.h"
	#include "library/vc/deditec_timer.h"
	#include "library/vc/deditec_string_utils.h"
	#include "library/delib/delib_error_codes.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
	#include "library/communication/dt_communication_paramams.h"
	#include "library/embedded/deditec_debug_comfort.h"
	
	#include "delib_packet_statistic.h"
	#include "delib_io_ftdi.h"
	#include "delib_hdlmgr.h"
	#include "delib_set_get_error.h"
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include <windows.h>
	#include "conio.h"
	#include "malloc.h"
	#include "delib_io_ftdi.h"
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
#endif

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ------------------------------------
// ------------------------------------
// Global Vars
// ------------------------------------
// ------------------------------------

ULONG second_call_prevent = 0;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiOpenModuleIntern_FTDI(ULONG handle32, ULONG moduleID, ULONG moduleNr, ULONG BusType, ULONG retry_max, ULONG subdevice, FT_HANDLE* ft_handle, ULONG open_options)
{
	DAPI_HANDLE *DapiHandle;
	ULONG ret;
	ULONG retry;
	char reg_key_name[30];
	char reg_key_param[40];
	ULONG err;

//	#ifdef DEBUG_OUTPUT
	char debug_msg[200];
//	#endif

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	ret = DapiOpenModuleGetFT_HANDLE(handle32, moduleID, moduleNr, BusType, retry_max, subdevice, ft_handle, open_options);

	if(moduleID == USB_CANOPEN_STICK)
	{
		DapiHandle->CANopenStickSondermodus = 1;
	}
	else
	{
		DapiHandle->CANopenStickSondermodus = 0;
	}

	sprintf(debug_msg,"CANopenStickSondermodus = %d\n", DapiHandle->CANopenStickSondermodus);
	debug_info(debug_msg);


	if(open_options == DAPI_OPEN_MODULE_OPTION_CANOPENSTICK_NO_REGISTER_REROUTE)
	{
		DapiHandle->CANopenStickSondermodus = 0;

		sprintf(debug_msg,"CANopenStickSondermodus = %d (open_options == DAPI_OPEN_MODULE_OPTION_CANOPENSTICK_NO_REGISTER_REROUTE)\n", DapiHandle->CANopenStickSondermodus);
		debug_info(debug_msg);
	}

	#ifdef __linux__
		// LINUX ---------------
		// LINUX ---------------
		// LINUX ---------------
		
		// bei linux ist der retry per default auf 5
		retry = 1;
		
		// LINUX ---------------
		// LINUX ---------------
		// LINUX ---------------		
	#else
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
		
		// bei windows wird der retry aus der win-registry gelesen
		
		// --------------------------------------------------------------------
		strcpy_n(reg_key_name, sizeof(reg_key_name), "DELIB-CONFIG\\USB\\Config", (ULONG) strlen("DELIB-CONFIG\\USB\\Config"));
		strcpy_n(reg_key_param, sizeof(reg_key_param), "Retry", (ULONG) strlen("Retry"));
		// --------------------------------------------------------------------
		//
		// ------------------------------------------------
		err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &retry);
		if (err == RETURN_OK)
		{
			// DEBUG
			#ifdef DEBUG_OUTPUT
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, retry);
				debug_info(debug_msg);
			#endif
			// -----
		}
		else
		{
			// DEBUG
			#ifdef DEBUG_OUTPUT
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
				debug_info(debug_msg);
			#endif
			// -----

			retry = 5;
			debug_error("Parameter not found in Registry: Using default Value: retry=5");
		}
		// ------------------------------------------------

		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	#endif
	
	DapiHandle->retry = retry;
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

int IsFTDIError(FT_STATUS dStatus)
{
	if(dStatus != FT_OK)
	{
		return (TRUE);
	}
	else
	{
		return (FALSE);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiFTDIdoReConnect(ULONG handle32)
{
	FT_HANDLE ftHandle;
	DAPI_HANDLE *DapiHandle;
	ULONG ret;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return;
	}

	#ifdef __linux__
		// LINUX ---------------
		// LINUX ---------------
		// LINUX ---------------
		return;
	#else
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
		
		// DEBUG
		char debug_msg[200];
		debug_set_function_name("DapiFTDIdoReConnect");
		sprintf(debug_msg, "Searching module... ");
		debug_info(debug_msg);
		// -----

		// Try ONE reconnect
		if(DapiHandle->ftHandle != 0)
		{
			// DEBUG
			debug_set_function_name("DapiFTDIdoReConnect");
			sprintf(debug_msg, "FT_Close ... ftHandle=0x%x", DapiHandle->ftHandle);
			debug_collect(debug_msg);
			// -----

			FT_Close(DapiHandle->ftHandle);
		}

		// DEBUG
		debug_set_function_name("DapiFTDIdoReConnect");
		sprintf(debug_msg, "DapiOpenModuleGetFT_HANDLE ... ");
		debug_collect(debug_msg);
		// -----

		ret = DapiOpenModuleGetFT_HANDLE(handle32, DapiHandle->moduleID, DapiHandle->moduleNr, DapiHandle->BusType, 10, DapiHandle->SubDevice, &ftHandle, 0);		// try to open / retry=10

		// DEBUG
		debug_set_function_name("DapiFTDIdoReConnect");
		sprintf(debug_msg, "DapiOpenModuleGetFT_HANDLE -> new ftHandle=0x%x - ret=%d", ftHandle, ret);
		debug_collect(debug_msg);
		// -----

		DapiHandle->ftHandle = ftHandle;	// Modul hat jetzt anderes Handle
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifndef USE_USB_SIMPLE
	ULONG DapiFTDISetModuleAdress(ULONG handle32, ULONG moduleNr_new)
	{
		unsigned char buffer[20];
		FT_STATUS ftStatus;
		char debug_msg[200];
		ULONG ret;
		DAPI_HANDLE *DapiHandle;
		ULONGLONG handle_ull;

		if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
		{
			DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
		}
		else
		{
			return DAPI_ERR_GEN_ILLEGAL_HANDLE;
		}

		if(moduleNr_new >= MAX_NO_OF_MODULES_NR) moduleNr_new = MAX_NO_OF_MODULES_NR-1;

		buffer[0] = 0x7a;													// Check ID for DEDITEC Entry
		buffer[1] = 0x3f;													// Check ID for DEDITEC Entry
		buffer[2] = (unsigned char) (DapiHandle->moduleID>>8) & 255;		// DELIB ModuleID_H
		buffer[3] = (unsigned char) (DapiHandle->moduleID     &255);		// DELIB ModuleID_L
		buffer[4] = (unsigned char) moduleNr_new;							// DELIB ModuleNr
		buffer[5] = 0x2f;													// Check ID for DEDITEC Entry

						
		ftStatus = FT_EE_UAWrite(DapiHandle->ftHandle, (unsigned char*)buffer, 12); 
		if(ftStatus != FT_OK) 
		{ 
			// DEBUG
			debug_set_function_name("DapiFTDISetModuleAdress");
			sprintf(debug_msg, "OpenModule FT_EE_UAWrite FAILED! ");
			debug_error(debug_msg);
			// -----
			
			ret=-1;	// Failed
		} 
		else 
		{ 
			// DEBUG
			debug_set_function_name("DapiFTDISetModuleAdress");
			sprintf(debug_msg, "OpenModule FT_EE_UAWrite OK! - data=0x%2x 0x%2x 0x%2x 0x%2x 0x%2x 0x%2x ", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5]);
			debug_info(debug_msg);
			// -----
		
			ret=0;	// OK
		}

		return ret;
	}
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifdef USE_FTDI
ULONG DapiOpenModuleGetFT_HANDLE(ULONG handle32, ULONG moduleID, ULONG moduleNr, ULONG BusType, ULONG retry_max, ULONG subdevice, FT_HANDLE* ret_ft_handle, ULONG open_options)
{
	char msg[200];
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	unsigned long NumDevices = 0;
	ULONG LocID = 0, ChipID = 0;
	FT_STATUS dStatus;
	FT_HANDLE ftHandle;
	FT_STATUS ftStatus;
	DWORD iNumDevs;
	long dwLoc;
	ULONG sn;
	char text_dev_nr[16];
	unsigned char buffer[64];
	DWORD BytesRead;
	ULONG i;
	FT_DEVICE_LIST_INFO_NODE *devInfo;
	ULONG check_device;
	char debug_msg[200];
	DWORD dwDriverVer;
	FT_STATUS status;
	ULONG ModuleLocation;
	ULONG baudrate;
	ULONG ok;

	ULONG pos_mid_b00_07 = 0;	// buffpos moduleID Bit00..07 in devInfo[check_device].SerialNumber[] 
	ULONG pos_mid_b08_15 = 0;	// buffpos moduleID Bit08..15 in devInfo[check_device].SerialNumber[] 
	ULONG pos_mid_b16_23 = 0;	// buffpos moduleID Bit16..23 in devInfo[check_device].SerialNumber[] 
	ULONG pos_ft_ab = 0;		// buffpos FTDI converter type (A/B)
	bool new_sn = FALSE;

	if(subdevice==0)
	{
		sprintf(text_dev_nr , "DT00%4dA", (int) moduleID);
		#ifdef USE_USB_SIMPLE
			// Hier wird nur noch mit dem "neuen" Seriannummernformat gearbeitet
			sprintf(text_dev_nr , "DT%03dA", (int) moduleID);
		#endif

	}
	else if(subdevice==1)
	{
		sprintf(text_dev_nr , "DT00%4dB", (int) moduleID);
		#ifdef USE_USB_SIMPLE
			// Hier wird nur noch mit dem "neuen" Seriannummernformat gearbeitet
			sprintf(text_dev_nr , "DT%03dB", (int) moduleID);
		#endif
	}
	else
	{
		// Ohne A oder B
		sprintf(text_dev_nr , "DT00%4d", (int) moduleID);
		#ifdef USE_USB_SIMPLE
			// Hier wird nur noch mit dem "neuen" Seriannummernformat gearbeitet
			sprintf(text_dev_nr , "DT%03d", (int) moduleID);
		#endif
	}

	// DEBUG
	debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
	sprintf(debug_msg, "DapiOpenModuleGetFT_HANDLE(0x%x, 0x%x, 0x%x, %s) Normal - open_options=0x%lx", (unsigned int) moduleID, (unsigned int) moduleNr, (unsigned int) retry_max, subdevice==DapiModuleID_Device?"NormalDevice":"ProgDevice", open_options);
	debug_info(debug_msg);
	// -----
	
	for(i=4;i!=8;++i)
	{
		if(text_dev_nr[i]<'0' || text_dev_nr[i]>'9') text_dev_nr[i]='0';
	}
	text_dev_nr[9]=0;

	// DEBUG
	debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
	sprintf(debug_msg, "Searching for the following S/Ns = %s / DT%03dxxxxx", text_dev_nr, (int) moduleID);
	debug_info(debug_msg);
	// -----

	// Reset LocID
	ModuleLocation = -1;

	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// NOT USE_USB_SIMPLE
	#ifndef USE_USB_SIMPLE

	ok=0;

	if ((open_options & DAPI_OPEN_MODULE_OPTION_NO_RESCAN) == 0)
	{
		// Open with Scan
		ftStatus = FT_CreateDeviceInfoList(&iNumDevs);
		if (ftStatus != FT_OK)
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Module Location = 0 - Open Error");
			debug_error(debug_msg);
			// -----
			
			*ret_ft_handle = (FT_HANDLE) -1;
			return DAPI_ERR_COM_FT_ERROR;
		}

		devInfo = (FT_DEVICE_LIST_INFO_NODE*) malloc(sizeof(FT_DEVICE_LIST_INFO_NODE) * iNumDevs);

		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "Number of connected devices = %d", (unsigned int) iNumDevs);
		debug_info(debug_msg);
		sprintf(debug_msg, "FT_GetDeviceInfoDetail - now");
		debug_info(debug_msg);
		// -----
		
		ftStatus = FT_GetDeviceInfoList(devInfo, &iNumDevs);
		if(ftStatus != FT_OK)
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Module Location = 0 - Open Error");
			debug_error(debug_msg);
			// -----

			free(devInfo);
			
			*ret_ft_handle = (FT_HANDLE) -1;
			return DAPI_ERR_COM_FT_ERROR;
		}



		ok=0;

		for(check_device=0;(check_device < iNumDevs) && (ok != 2);++check_device)
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Dev Nr %d SerialNumber = %s", (unsigned int) check_device, devInfo[check_device].SerialNumber);debug_info(debug_msg);
			sprintf(debug_msg, "Dev Nr %d ftHandle= %lx",(unsigned int) check_device, (unsigned long) devInfo[check_device].ftHandle);debug_info(debug_msg);
			sprintf(debug_msg, "Dev Nr %d Location= %d",(unsigned int) check_device, (unsigned int) devInfo[check_device].LocId);debug_info(debug_msg);
			// -----

			if ((devInfo[check_device].SerialNumber[2] == '0') && (devInfo[check_device].SerialNumber[3] == '0') && (devInfo[check_device].SerialNumber[4] == '0'))
			{
				// ALT
				// S/N = DT000xxxy		xxx=ModuleID	y=0/'A'/'B'
				pos_ft_ab = 8;
				pos_mid_b00_07 = 7;
				pos_mid_b08_15 = 6;
				pos_mid_b16_23 = 5;
				new_sn = FALSE;
			}
			else
			{
				// NEU
				// S/N = DTxxx0yyyyyyz		xxx=ModuleID + zzzz=auto-generate + y='0'/'A'/'B'
				//							DT0240 (id) + 22EAIE (auto) + 0/'A'/'B 
				// Beispiel RO-CPU-CAN2 (ohne A/B)		"DT024022EAIE"
				// Beispiel BS-USB-OR8  (mit  A/B)		"DT01502VU8FXA"

				pos_ft_ab = 12;
				pos_mid_b00_07 = 4;
				pos_mid_b08_15 = 3;
				pos_mid_b16_23 = 2;
				new_sn = TRUE;
			}

			sn = (asc2hex(devInfo[check_device].SerialNumber[pos_mid_b00_07]) & 15);
			sn += (asc2hex(devInfo[check_device].SerialNumber[pos_mid_b08_15]) & 15) * 10;
			sn += (asc2hex(devInfo[check_device].SerialNumber[pos_mid_b16_23]) & 15) * 100;
			//sn += (asc2hex(devInfo[check_device].SerialNumber[4]) & 15) *1000;

			dwLoc = devInfo[check_device].LocId;

			if(dwLoc != 0)
			{
				// Location != 0 ...... ber die location ffnen !
				dStatus = FT_OpenEx((void*) dwLoc, FT_OPEN_BY_LOCATION, &ftHandle);
		
				sprintf(debug_msg, "FT_OpenEx opens Module (FT_OPEN_BY_LOCATION) Loc=%ld", dwLoc);
				debug_info(debug_msg);
			}
			else
			{
				// Location == 0 ...... ber die Seriennummer ffnen !
				if (new_sn == TRUE)
				{
					// 06.06.18 CC spezial fr neue S/N
					if ((devInfo[check_device].SerialNumber[0] == 'D') && \
						(devInfo[check_device].SerialNumber[1] == 'T') && \
						((devInfo[check_device].SerialNumber[2] >= '0') && (devInfo[check_device].SerialNumber[2] <= '9')) && \
						((devInfo[check_device].SerialNumber[3] >= '0') && (devInfo[check_device].SerialNumber[3] <= '9')) && \
						((devInfo[check_device].SerialNumber[4] >= '0') && (devInfo[check_device].SerialNumber[4] <= '9')))
					{
						// 06.06.18 CC nur kopieren, wenn es auch ein DEDITEC Modul ist..
						strcpy_n(text_dev_nr, sizeof(text_dev_nr), devInfo[check_device].SerialNumber, strlen_n(devInfo[check_device].SerialNumber, -1));
					}
				}

				dStatus = FT_OpenEx((void*) text_dev_nr, FT_OPEN_BY_SERIAL_NUMBER, &ftHandle);

				sprintf(debug_msg, "FT_OpenEx opens Module (FT_OPEN_BY_SERIAL_NUMBER) S/N = %s", text_dev_nr);
				debug_info(debug_msg);
			}

			if(dStatus != FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "Open Error");
				debug_error(debug_msg);
				// -----
				
				*ret_ft_handle = (FT_HANDLE) -1;
			}
			else
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "dStatus=FT_OK * Location= %d", (unsigned int) dwLoc);
				debug_info(debug_msg);
				// -----
				
				if(moduleID == 0)
				{
					// This is the searched Module
					ModuleLocation = dwLoc;
				
					// DEBUG
					debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
					sprintf(debug_msg, "ModuleID = 0 -> Ignore EEPROM Settings");
					debug_info(debug_msg);
					// -----
				}
				else if((sn != moduleID))
				{
					// DEBUG
					debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
					sprintf(debug_msg, "Serial No does not match moduleID=%x, sn=%x, serial-No = %s", (unsigned int) moduleID, (unsigned int) sn, devInfo[check_device].SerialNumber);
					debug_info(debug_msg);
					// -----
				}
				else
				{
					// Check Module-Nr

					ok=0;

					if((subdevice==0) || (subdevice==1))
					{	
						// Subdevice must be checked too
						if(	(subdevice==0) && (devInfo[check_device].SerialNumber[pos_ft_ab]=='A') || \
							(subdevice==1) && (devInfo[check_device].SerialNumber[pos_ft_ab]=='B'))
						{
							// This is the searched Module
							ModuleLocation = dwLoc;
					
							ok=1;

							// DEBUG
							debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
							sprintf(debug_msg, "User Bytes : correct subdevice found");
							debug_info(debug_msg);
							// -----
						}
						else
						{
							// DEBUG
							debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
							sprintf(debug_msg, "User Bytes : Subdevice = %d - SerialNumber[%ld]=%c does not match", (unsigned int) subdevice, pos_ft_ab, devInfo[check_device].SerialNumber[pos_ft_ab]);
							debug_info(debug_msg);
							// -----
						}
					}
					else
					{
						// This is the searched Module
						ok=1;
						sprintf(debug_msg,"OpenModule User Bytes : correct ModuleNr found");
						debug_print(debug_msg);
					}
					
					if(ok==1)
					{
						// Subdevice-Nr stimmt - Jetzt EEPROM checken
						ftStatus = FT_EE_UARead(ftHandle, buffer, 64, &BytesRead);
						if (ftStatus == FT_OK) 
						{
							// DEBUG
							debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
							sprintf(debug_msg, "%ld User Bytes read: %x %x %x %x %x %x %x %x ", BytesRead, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
							debug_info(debug_msg);
							// -----

							if((buffer[0] == 0x7a) && (buffer[1] == 0x3f) && (buffer[5] == 0x2f))
							{
								// DEBUG
								debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
								sprintf(debug_msg, "User Bytes : correct DEDITEC ID found");
								debug_info(debug_msg);
								// -----

								ULONG moduleIDXXX = (((unsigned long) buffer[2]) << 8) |  ((unsigned long) buffer[3]);

								if(moduleID == moduleIDXXX)
								{
									// DEBUG
									debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
									sprintf(debug_msg, "User Bytes : correct ModuleID found");
									debug_info(debug_msg);
									sprintf(debug_msg, "buffer[4]=0x%x  moduleNr&255=0x%x", buffer[4], (unsigned int) (moduleNr&255));
									debug_info(debug_msg);
									// -----
									
									if(buffer[4] == (moduleNr&255))
									{	
										//OK
										ok=2;
										ModuleLocation = dwLoc;

										// DEBUG
										debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
										sprintf(debug_msg, "ModuleLocation=%d ok=2", (unsigned int) ModuleLocation);
										debug_info(debug_msg);
										// -----
									
										if(sn != moduleIDXXX)
										{
											debug_set_function_name("DapiScanAllModulesAvailable");
											sprintf(msg, "******* sn != moduleID *******   ERROR Fxxx Driver Problem");
											debug_error(msg);
										}

									}
									else
									{
										// DEBUG
										debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
										sprintf(debug_msg, "User Bytes : ModuleNr =0x%x (does not match 0x%x)", buffer[4], (unsigned int) moduleNr);
										debug_info(debug_msg);
										// -----
									}
								}
								else
								{
									// DEBUG
									debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
									sprintf(debug_msg, "User Bytes : ModuleID =0x%x (does not match 0x%x)", buffer[2]<<8 | buffer[3] , (unsigned int) moduleID);
									debug_info(debug_msg);
									// -----
								}	
							}
						} 
						else 
						{ 
							// FT_EE_UARead FAILED! 
							// DEBUG
							debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
							sprintf(debug_msg, "OpenModule FT_EE_UARead FAILED! ");
							debug_error(debug_msg);
							// -----
						}
					}
				}
				
				FT_Close(ftHandle);
				
			}

			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "ok=%d", (unsigned int) ok);
			debug_info(debug_msg);
			// -----

			if(ok==2)
			{
				// Richtiges Modul gefunden
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ALLES OK - ModuleLocation = %d", (unsigned int) dwLoc);
				debug_info(debug_msg);
				// -----
			}
			else
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "not ok-----------------------------------------------");
				debug_error(debug_msg);
				// -----
			}
			
		} //	for(check_device=0;check_device<iNumDevs;++check_device)

		free(devInfo);
	} // if(open_options == 0)
#ifndef __linux__
	else if((open_options & DAPI_OPEN_MODULE_OPTION_NO_RESCAN) != 0)
	{
		// No Rescan .... Get LocID from Scanned List

		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");

		sprintf(debug_msg, "No rescan....Searching for ModuleID(ModuleNr) = %ld(%ld) in AllModulesAvailableList(Entries=%ld)", moduleID, moduleNr, AllModulesAvailableCount);
		debug_info(debug_msg);
		// -----

		ok=0;
		i=0;
		ULONG ModulesAvailableListSearchingFor = (moduleID | (((unsigned long) moduleNr) << 16));
		while((i<AllModulesAvailableCount) && (ok==0))
		{
			if(AllModulesAvailableList[i] == ModulesAvailableListSearchingFor)
			{
				ok=2;
				ModuleLocation = AllModulesAvailableLocID[i];
			}
	
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "AllModulesAvailableList[%ld]=0x%lx   != 0x%lx ok = %ld", i, AllModulesAvailableList[i], ModulesAvailableListSearchingFor, ok);
			debug_info(debug_msg);
			// -----
	
			++i;
		}
	} // if(open_options == DAPI_OPEN_MODULE_OPTION_NO_RESCAN)
#endif
	
	if(ok==2)
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "FT_OpenEx(%d) will open now the Module. ModuleLocation=0x%x", (unsigned int) ModuleLocation,  (unsigned int) ModuleLocation);
		debug_info(debug_msg);
		// -----

		// open Module
		if(ModuleLocation != 0)
		{
			// Location != 0 ...... ber die location ffnen !
			dStatus = FT_OpenEx((void*) ModuleLocation, FT_OPEN_BY_LOCATION, &ftHandle);

			sprintf(debug_msg, "FT_OpenEx opens Module (FT_OPEN_BY_LOCATION) Loc=%ld", ModuleLocation);
			debug_info(debug_msg);
		}
		else
		{
			// Location == 0 ...... ber die Seriennummer ffnen !
			dStatus = FT_OpenEx((void*) text_dev_nr, FT_OPEN_BY_SERIAL_NUMBER, &ftHandle);

			sprintf(debug_msg, "FT_OpenEx opens Module (FT_OPEN_BY_SERIAL_NUMBER) S/N = %s", text_dev_nr);
			debug_info(debug_msg);
		}

		if(dStatus != FT_OK)
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Open Error");
			debug_error(debug_msg);
			// -----
			
			*ret_ft_handle = (FT_HANDLE) -1;
		}
		else
		{
			status = FT_GetDriverVersion(ftHandle, &dwDriverVer);
			if(status==FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "GetDriverVersion = %x", (unsigned int) dwDriverVer);
				debug_info(debug_msg);
				// -----
			}
			
			*ret_ft_handle = ftHandle;
		}
	}
	else
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "Module Location = 0 - Open Error");
		debug_error(debug_msg);
		// -----
		
		*ret_ft_handle = (FT_HANDLE) -1;
	}

	// DEBUG
	debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
	sprintf(debug_msg, "FTHANDLE = 0x%lx", (unsigned long) ftHandle);
	debug_info(debug_msg);
	// -----

	if(*ret_ft_handle != ((FT_HANDLE) -1))
	{
		if((BusType == bus_SeriellFTDI)  ||
		   (BusType == bus_FTDI_SINGLE_SER) ||
		   (BusType == bus_FTDI_SINGLE_PAR))
		{
			//dStatus = FT_SetTimeouts(ftHandle, 100*5, 100*5);

			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Timeout ok");
			debug_info(debug_msg);
			// -----
		
			if(( moduleID != USB_MINI_STICK) &&
				(moduleID != USB_LOGI_18) &&
				(moduleID != USB_SPI_MON) &&
				(moduleID != USB_WATCHDOG) &&
				(moduleID != USB_OPTOIN_8))
			{
				baudrate = 115200;
			}
			else
			{
				baudrate = (unsigned long) (250000.0 * 0.6);				// 250kBaud (Anpassung R81C 12 Mhz zu 20 MHz)
			}
			
			ftStatus=FT_SetBaudRate(ftHandle, baudrate);
			if (ftStatus != FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetBaudRate");
				debug_error(debug_msg);
				// -----
			}
			else
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "Serial Baudrate = %d", (unsigned int) baudrate);
				debug_info(debug_msg);
				// -----
			}
			// ----------------------------
			ftStatus=FT_SetDataCharacteristics(ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
			if (ftStatus != FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetDataCharacteristics");
				debug_error(debug_msg);
				// -----
			}
			// ----------------------------
			// ----------------------------
			ftStatus=FT_SetLatencyTimer(ftHandle,2); 
			if (ftStatus != FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetLatencyTimer");
				debug_error(debug_msg);
				// -----
			}
		// ----------------------------
		}
		// Diese Einstellungen sind fr jedes FTDI-Device
		// ----------------------------
		ftStatus=FT_SetTimeouts(ftHandle,1,50); 
		if (ftStatus != FT_OK)
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetTimeouts");
			debug_error(debug_msg);
			// -----
		}
		// ----------------------------
		// ----------------------------



	}
	#endif


	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// USE_USB_SIMPLE - ONLY used for linux


	#ifdef USE_USB_SIMPLE
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	
	ftStatus = FT_OpenEx(text_dev_nr, FT_OPEN_BY_SERIAL_NUMBER, ret_ft_handle, BusType);
	if (ftStatus == FT_OK)
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "FTDI-Handle = 0x%lx", (unsigned long) *ret_ft_handle);
		debug_info(debug_msg);
		// -----
	}
	else
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "FTDI OPEN ERROR - text_dev_nr = %s\n\r", text_dev_nr);
		debug_error(debug_msg);
		// -----
		
		*ret_ft_handle = (FT_HANDLE) -1;
	}

	if(*ret_ft_handle != ((FT_HANDLE) -1))
	{
		if((BusType == bus_SeriellFTDI)  ||
		   (BusType == bus_FTDI_SINGLE_PAR) ||
		   (BusType == bus_FTDI_SINGLE_SER))
		{
			// DEBUG
			debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
			sprintf(debug_msg, "Timeout ok");
			debug_info(debug_msg);
			// -----
		
			if((moduleID != USB_MINI_STICK) && (moduleID != USB_LOGI_18) && (moduleID != USB_SPI_MON) && (moduleID != USB_WATCHDOG) && (moduleID != USB_OPTOIN_8))
			{
				baudrate = 115200;
			}
			else
			{
				baudrate = (unsigned long) (250000/1.666);					// 150 KBaud beim R8C1A
			}
			// ----------------------------
			if(ftdi_set_baudrate ((struct ftdi_context *) *ret_ft_handle, baudrate) != 0)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetBaudRate");
				debug_error(debug_msg);
				// -----
			}
			else
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "Serial Baudrate = %d", (unsigned int) baudrate);
				debug_info(debug_msg);
				// -----
			}
			// ----------------------------
			if(ftdi_set_latency_timer((struct ftdi_context *) *ret_ft_handle, 2) != 0)
			{
				// DEBUG
				debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
				sprintf(debug_msg, "ftStatus != FT_OK -> FT_SetLatencyTimer");
				debug_error(debug_msg);
				// -----
			}
			// ----------------------------
		}
	}
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	#endif 


	if(*ret_ft_handle == ((FT_HANDLE) -1)) 
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "END *******   return = -1 (ERROR) -------------------------------------");
		debug_info(debug_msg);
		// -----

		return DAPI_ERR_COM_FT_ERROR;
	}
	else 
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleGetFT_HANDLE");
		sprintf(debug_msg, "END *******   return = 0 (ok)  -------------------------------------");
		debug_info(debug_msg);
		// -----

		return DAPI_ERR_NONE;
	}
}

#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG FTread_n_bytes_with_timeout(ULONG handle32, ULONG anz, unsigned char *buff, ULONG delay_max)
{
	//ULONG buff[1000];
	ULONG rx_pos;
	DWORD rx_count;
	FT_STATUS  dStatus;
	char debug_msg[200];
	ULONG i;
	ULONG exit;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	rx_pos=0;
	exit=0;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("FTread_n_bytes_with_timeout");
		sprintf(debug_msg, "delay_max = %ld", delay_max);
		debug_error(debug_msg);
	#endif
	// -----

	dt_timerus_start();
	do 
	{
		dStatus = FT_Read(DapiHandle->ftHandle, buff+rx_pos, anz-rx_pos, &rx_count);

		if (dStatus != FT_OK)
		{
			// DEBUG
			debug_set_function_name("FTread_n_bytes_with_timeout");
			sprintf(debug_msg, "dStatus != FT_OK -> FT_Read");
			debug_error(debug_msg);
			// -----
		}
		
		#ifdef DEBUG_OUTPUT
			debug_set_function_name("FTread_n_bytes_with_timeout");
			sprintf(debug_msg, "anz=%ld, rx_pos=%ld", anz, rx_pos);
			debug_print(debug_msg);
		#endif

		rx_pos+=rx_count;

	} while((rx_pos<anz) && (dt_timerus_diff()<delay_max) && (exit==0));

	i=dt_timerus_diff();

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("FTread_n_bytes_with_timeout");
		sprintf(debug_msg, "delay = %ld    anz=%ld, rx_pos=%ld", i, anz, rx_pos);
		debug_print(debug_msg);
	#endif

	if(rx_pos<anz) 
	{
		return rx_pos;		// received not enough bytes
	}
	else 
	{
		return anz;
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDISendCommand(ULONG handle32, ULONG command, ULONG adress, ULONG be, UCHAR *data_tx, ULONG data_tx_length, UCHAR *data_rx, ULONG data_rx_length)
{
	DAPI_HANDLE *DapiHandle;
	FT_STATUS dStatus;
	char debug_msg[200];
	unsigned char buff[10000];
	ULONG buff_cnt;
	DWORD buff_cnt_ret;
	ULONG retry_init_module;
	ULONG i;
	ULONG cnt=0;
	ULONG get_char_cnt=0;
	ULONG comm_error=0;
	ULONG retry_cnt;
	ULONG module_error_code = DAPI_ERR_NONE;
	ULONG ret;
	ULONG checksum_calc;
	ULONG checksum_rx;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// DEBUG
	debug_collect_init();

	debug_set_function_name("DapiFTDISendCommand");
	
	sprintf(debug_msg, "DapiFTDISendCommand START");
	debug_collect(debug_msg);
	
	sprintf(debug_msg, "DapiFTDISendCommand(handle=0x%lx, command=0x%lx, address=0x%lx, be=0x%lx, data_tx_length=0x%lx, data_rx_length=0x%lx)", handle32, command, adress, be, data_tx_length, data_rx_length);
	debug_collect(debug_msg);
	// -----
	
	
	if(second_call_prevent != 0 ) 
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "Second call in background detected - STOP it");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return DAPI_ERR_COM_DELIB_ID_INVALID;
	}

	if (handle32 == 0)
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// 07.04.17	CC	Aenderung fr 64bit Linux - ohne Cast ist hier IMMER true, weil ULONG hier 8bytes sind
	if (((unsigned int)DapiHandle->id) != ((unsigned int)delib_id))
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_COM_DELIB_ID_INVALID;
	}

	++second_call_prevent;	// prevent a background call 
	++DapiHandle->jobID;
	DapiHandle->jobID &= 0xff;

	retry_cnt = 0;

	// DEBUG
	debug_set_function_name("DapiFTDISendCommand");
	sprintf(debug_msg, "Start with retries Nr %ld DapiHandle->ftHandle=0x%lx", retry_cnt, (ULONG) DapiHandle->ftHandle);
	debug_collect(debug_msg);
	// -----
	
	do 
	{
		if(retry_cnt > 0)
		{
			++DapiHandle->packet_statistic_transmit_normal_count_retry;
			// DEBUG
			debug_set_function_name("DapiFTDISendCommand");
			sprintf(debug_msg, "ERROR aufgetreten-----------------------------------------");
			debug_collect(debug_msg);
			sprintf(debug_msg, "RETRY RETRY RETRY RETRY = %ld / %ld", (unsigned long) retry_cnt, (unsigned long) 2*DapiHandle->retry);
			debug_collect(debug_msg);
			// -----


			if(retry_cnt == DapiHandle->retry)
			{
				// Sonderfall - mittendrin ein ReConnect durchfhren
				DapiFTDIdoReConnect(handle32);
				++DapiHandle->packet_statistic_transmit_normal_count_recon;
			}
			else
			{
				// Empfangsbuffer flushen
				ret = FTread_n_bytes_with_timeout(handle32, 200, buff, 50*1000);	// Read x Byte (max 50ms->zum flushen)
			}

			// DEBUG
			debug_set_function_name("DapiFTDISendCommand");
			sprintf(debug_msg, "FLUSH BUFFER - buff_cnt_ret = %ld bytes wurden geflusht", ret);
			debug_collect(debug_msg);
			if (ret > 0)
			{
				for (i=0; i<ret; i+=16)
				{
					debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), buff, ret, i, ret);
					debug_collect(debug_msg);
				}
			}
			// -----
		}

		comm_error=0;

		buff_cnt=0;
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 
		buff[buff_cnt++] = 0xfe;			// Dummy Zeichen 

	
		buff[buff_cnt++] = 0xff;
		buff[buff_cnt++] = (unsigned char) DapiHandle->jobID;

		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			buff[buff_cnt++] = (unsigned char) (command | 0x80);		// Kennzeichne "neues Protokoll"
		}
		else
		{
			buff[buff_cnt++] = (unsigned char) command;
		}

		buff[buff_cnt++] = (unsigned char) ((adress      ) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >>  8) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >> 16) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >> 24) & 0xff);

		buff[buff_cnt++] = (unsigned char) (be & 0xff);
		// das wird IMMER gesendet

		checksum_calc = 0;
		checksum_calc += (unsigned char) buff[buff_cnt - 7];
		checksum_calc += (unsigned char) buff[buff_cnt - 6];
		checksum_calc += (unsigned char) buff[buff_cnt - 5];
		checksum_calc += (unsigned char) buff[buff_cnt - 4];
		checksum_calc += (unsigned char) buff[buff_cnt - 3];
		checksum_calc += (unsigned char) buff[buff_cnt - 2];
		checksum_calc += (unsigned char) buff[buff_cnt - 1];

		if(command == DT_FTDI_USB_CMD_SEND8BYTE)
		{
			// always send 8 bytes
			for(i=0;i!=8;++i)
			{
				if(i < data_tx_length) 
				{
					buff[buff_cnt++] = *(data_tx + i);
					checksum_calc += *(data_tx + i);
				}
				else
				{
					buff[buff_cnt++] = 0;
				}
			}
		}

		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			buff[buff_cnt++] = (unsigned char) (checksum_calc       & 0xff);			// Checksum 
			buff[buff_cnt++] = (unsigned char) ((checksum_calc >> 8) & 0xff);			// Checksum 
		}

		retry_init_module=5;

		do
		{
			dStatus = FT_Write(DapiHandle->ftHandle, buff, buff_cnt, &buff_cnt_ret);
			if(dStatus != FT_OK)
			{
				// DEBUG
				debug_set_function_name("DapiFTDISendCommand");
				sprintf(debug_msg, "dStatus != FTDI_SUCCESS   dStatus = %x", (unsigned int) dStatus);
				debug_collect(debug_msg);
				// -----

				DapiFTDIdoReConnect(handle32);			// Try reconnect
			}
			--retry_init_module;
		} while( (dStatus != FT_OK) && (retry_init_module > 0) );
	
		if(retry_init_module == 0)
		{
			--second_call_prevent;	// prevent a background call 
			return DAPI_ERR_COM_FT_ERROR;
		}

		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "Sending %ld Bytes", buff_cnt);
		debug_collect(debug_msg);
		if (buff_cnt > 0)
		{
			for (i=0; i<buff_cnt; i+=16)
			{
				debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), buff, sizeof(buff), i, buff_cnt);
				debug_collect(debug_msg);
			}
		}

		if(command == DT_FTDI_USB_CMD_PING) 
		{
			get_char_cnt=7;		// Start_char + job_id + 4bytes-data + end_char
		}
	
		if(command == DT_FTDI_USB_CMD_SEND8BYTE) 
		{
			get_char_cnt=3;	// Start_char + job_id + 0bytes-data + end_char
		}

		if(command == DT_FTDI_USB_CMD_GET8BYTE) 
		{
			get_char_cnt=11;	// Start_char + job_id + 8bytes-data + end_char
		}

		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			get_char_cnt += 4; // 4 Zeichen bei neuem Protokoll mehr !
		}

		buff_cnt_ret = FTread_n_bytes_with_timeout(handle32, get_char_cnt, buff, 200*1000);	// max 100ms // 200ms - Win7 Probleme Fix. 6/4/2011
		// buff_cnt_ret = n , es wurden "n" Bytes empfangen

		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "printing received %ld bytes (sollten sein = %ld)", (unsigned long) buff_cnt_ret, (unsigned long) get_char_cnt);
		debug_collect(debug_msg);
		if(buff_cnt_ret > 0)
		{
			for (i=0; i<buff_cnt_ret; i+=16)
			{
				debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), buff, sizeof(buff), i, buff_cnt_ret);
				debug_collect(debug_msg);
			}
		}
		// -----

		if(buff_cnt_ret != get_char_cnt) 
		{
			// DEBUG
			debug_set_function_name("DapiFTDISendCommand");
			sprintf(debug_msg, "COMM-ERROR: ft_rx ----- received char count = %d (sollte sein = %d)", (unsigned int) buff_cnt_ret, (unsigned int) get_char_cnt);
			debug_collect(debug_msg);
			// -----
			
			++comm_error;
		}
		else
		{
			for(i=0;i!=8;++i) 
			{
				if(i<data_rx_length)
				{
					*(data_rx + i) = buff[i+2];
				}
			}

			checksum_calc = 0;
			// Checksumme bezieht sich nicht auf die letzten 3 Zeichen
			for(i=0;i!=(get_char_cnt-3);++i) 
			{
				checksum_calc += (unsigned char) buff[i];
			}

			if(buff[0] != 0x1A) 
			{
				// DEBUG
				debug_set_function_name("DapiFTDISendCommand");
				sprintf(debug_msg, "COMM-ERROR: ft_rx_buff[0]=%x  (sollte sein=%x)", (unsigned int) buff[0], 0x1a);
				debug_collect(debug_msg);
				// -----
				
				++comm_error;
			}
	
			if(buff[1] != DapiHandle->jobID) 
			{
				// DEBUG
				debug_set_function_name("DapiFTDISendCommand");
				sprintf(debug_msg, "COMM-ERROR: ft_rx_buff[1]=%x sollte sein=%x (JOB-ID)", (unsigned int) buff[1], (unsigned int) DapiHandle->jobID);
				debug_collect(debug_msg);
				// -----
				
				++comm_error;
			}

			if(DapiHandle->ftdi_use_protocol_new != TRUE)
			{
				// ALTES Protokoll
				if(buff[get_char_cnt-1] != 0x4f)
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommand");
					sprintf(debug_msg, "COMM-ERROR: ft_rx_buff[%d]=%x sollte sein=0x%x", (unsigned int) get_char_cnt-1, (unsigned int) buff[get_char_cnt-1], 0x4f);
					debug_collect(debug_msg);
					// -----
					
					++comm_error;
				}
			}
			else
			{
				// NEUES Protokoll
				// Check Zeichen: Ende altes Protokoll
				if(buff[get_char_cnt-5] != 0x4f)
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommand");
					sprintf(debug_msg, "COMM-ERROR: ft_rx_buff[%d]=%x sollte sein=0x%x", (unsigned int) get_char_cnt-4, (unsigned int) buff[get_char_cnt-4], 0x4f);
					debug_collect(debug_msg);
					// -----
					
					++comm_error;
				}

				// 1. Byte "neues" Protokoll // ERROR Code
				module_error_code = (buff[get_char_cnt-4] & 0xff);
				if(module_error_code != DAPI_ERR_NONE)
				{
					module_error_code |= DAPI_ERR_DEV_CLASS;
				}		

				// DEBUG			
				debug_set_function_name("DapiFTDISendCommand");
				sprintf(debug_msg, "Empfangener ERROR Code vom Modul = 0x%lx", module_error_code);
				debug_collect(debug_msg);
				// -----

				switch(module_error_code)
				{
					case DAPI_ERR_NONE:						break;
					case DAPI_ERR_DEV_IO_ADDR_ILLEGAL:		break;
					case DAPI_ERR_DEV_IO_OTHER_ERROR:		break;
					case DAPI_ERR_DEV_IO_RD_WR_ERROR:		break;
					case DAPI_ERR_DEV_IO_SUB_ADDR_ILLEGAL:	break;
					default:	++comm_error;				break;
				}

				// Check 2.+3. Byte "neues" Protokoll / CHECKSUM
				checksum_rx  =  (unsigned long) (buff[get_char_cnt-3] & 0xff);
				checksum_rx += ((unsigned long) (buff[get_char_cnt-2] & 0xff) << 8);
				if(checksum_rx != (checksum_calc & 0xffff))
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommand");
					sprintf(debug_msg, "COMM-ERROR: ft_rx_ checksum_rx=0x%lx sollte sein=0x%x (CHECKSUM ERROR)", checksum_rx, (unsigned int) (checksum_calc & 0xffff));
					debug_collect(debug_msg);
					// -----
					
					++comm_error;
				}


				// Check 4. Byte "neues" Protokoll
				if(buff[get_char_cnt-1] != 0xaf)
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommand");
					sprintf(debug_msg, "COMM-ERROR: ft_rx_buff[%d]=0x%x sollte sein=0x%x", (unsigned int) get_char_cnt-1, (unsigned int) buff[get_char_cnt-1], 0xaf);
					debug_collect(debug_msg);
					// -----
					
					++comm_error;
				}
			}
		}

		if(comm_error!=0)
		{			
			dStatus = FT_GetQueueStatus(DapiHandle->ftHandle, &buff_cnt_ret);
			if(IsFTDIError(dStatus)) 
			{
				--second_call_prevent;	// prevent a background call 
				
				// DEBUG
				debug_set_function_name("DapiFTDISendCommand");
				sprintf(debug_msg, "COM-ERROR!=0: ERROR FT_GetQueueStatus");
				debug_collect(debug_msg);
				debug_collect_printall();
				// -----
				
				return DAPI_ERR_COM_FT_ERROR;	// Fehler
			}

			// DEBUG
			debug_set_function_name("DapiFTDISendCommand");
			sprintf(debug_msg, "COM-ERROR!=0: Rx Zeichen noch im buffer da: %d", (unsigned int) buff_cnt_ret);
			debug_collect(debug_msg);
			// -----
		}


		/*
		#ifndef USE_USB_SIMPLE
			dStatus = FT_GetQueueStatus(DapiHandle->ftHandle, &buff_cnt_ret);
			if(buff_cnt_ret!=0)
			{
				FT_Purge( DapiHandle->ftHandle, FT_PURGE_RX);
				if(IsFTDIError(dStatus)) 
				{
					--second_call_prevent;	// prevent a background call 
					
					// DEBUG
					debug_set_function_name("DapiFTDISendCommand");
					sprintf(debug_msg, "ERROR FT_Purge");
					debug_collect(debug_msg);
					debug_collect_printall();
					// -----
					
					return DAPI_ERR_COM_FT_ERROR;	// Fehler
				}
			}
		#endif
		*/
		
		if(comm_error != 0)
		{
			++retry_cnt;
		}

	} while( (comm_error!=0) && (retry_cnt <= (2*DapiHandle->retry)));

	--second_call_prevent;	// prevent a background call 

	if(comm_error==0) 
	{	
		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");
		sprintf(debug_msg, "COM-ERROR=0: DapiFTDISendCommand END * module_error_code = 0x%lx * return = 0x%lx -", module_error_code, ret);
		debug_collect(debug_msg);
		// -----

		if(module_error_code != DAPI_ERR_NONE)
		{
			// DEBUG DIREKT-AUSGABE !!!!
			debug_set_function_name("DapiFTDISendCommand");
			sprintf(debug_msg, "Empfangener ERROR Code vom Modul = 0x%lx", module_error_code);
			debug_info(debug_msg);
			// -----
		}

		return module_error_code;
	}
	else 
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommand");

		sprintf(debug_msg, "COM-ERROR!=0: DapiFTDISendCommand END * return = DAPI_ERR_DEV_PKT_OTHER_ERROR");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_DEV_PKT_OTHER_ERROR;	// Fehler
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDISendCommandMultipleBytes(ULONG handle32, ULONG command, ULONG adress, ULONG adress_tiefe, ULONG wiederholungen, unsigned char* data, ULONG data_length)
{

	DAPI_HANDLE *DapiHandle;
	FT_STATUS dStatus;
	unsigned char * buff;
	ULONG buff_cnt;
	DWORD buff_cnt_ret;
	ULONG alloc_mem_size = 0;
	ULONG i;
	ULONG j;
	ULONG rd_offset;
	ULONG cnt=0;
	ULONG get_char_cnt=0;
	ULONG error=0;
	ULONG be=0;
	char debug_msg[200];
	ULONG ret;
	ULONG retry_cnt;
	ULONG module_error_code = DAPI_ERR_NONE;
	ULONG checksum_calc;
	ULONG checksum_rx;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	// DEBUG
	debug_collect_init();
	debug_set_function_name("DapiFTDISendCommandMultipleByt");
	sprintf(debug_msg, "DapiFTDISendCommandMultipleBytes START");
	debug_collect(debug_msg);
	sprintf(debug_msg, "DapiFTDISendCommandMultipleBytes(handle=0x%lx, command=0x%lx, address=0x%lx, adress_tiefe=0x%lx, wiederholungen=0x%lx, data_length=0x%lx)", handle32, command, adress, adress_tiefe, wiederholungen, data_length);
	debug_collect(debug_msg);
	// -----
	
	if(second_call_prevent != 0 ) 
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");

		sprintf(debug_msg, "Second call in background detected - STOP it");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_COM_DELIB_ID_INVALID;	// Fehler
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");

		sprintf(debug_msg, "Invalid Handle");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_COM_DELIB_ID_INVALID;	// Fehler
	}

	alloc_mem_size = (adress_tiefe+1)*wiederholungen + 100;

	// DEBUG
	debug_set_function_name("DapiFTDISendCommandMultipleByt");
	sprintf(debug_msg, "Allocating now memory (%d Bytes)", (unsigned int) alloc_mem_size);
	debug_collect(debug_msg);
	// -----

	buff = (unsigned char *) malloc (alloc_mem_size);

	if(buff == 0)
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "Allocating memory (%d Bytes) -> failure", (unsigned int) alloc_mem_size);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return DAPI_ERR_GEN_MALLOC_ERROR;	// Fehler
	}

	++second_call_prevent;	// prevent a background call 
	++DapiHandle->jobID;
	DapiHandle->jobID &= 0xff;
	
	retry_cnt = 0;
	
	// DEBUG
	debug_set_function_name("DapiFTDISendCommandMultipleByt");
	sprintf(debug_msg, "Start with retries Nr %ld DapiHandle->ftHandle=0x%lx", retry_cnt, (ULONG) DapiHandle->ftHandle);
	debug_collect(debug_msg);
	// -----
	
	do 
	{
		error=0;
		
		if(retry_cnt > 0)
		{
			++DapiHandle->packet_statistic_transmit_multiple_count_retry;

			// DEBUG
			debug_set_function_name("DapiFTDISendCommandMultipleByt");
			sprintf(debug_msg, "ERROR aufgetreten-----------------------------------------");
			debug_collect(debug_msg);
			sprintf(debug_msg, "RETRY RETRY RETRY RETRY = %ld / %ld", (unsigned long) retry_cnt, (unsigned long) 2*DapiHandle->retry);
			debug_collect(debug_msg);
			// -----

			if(retry_cnt == DapiHandle->retry)
			{
				// Sonderfall - mittendrin ein ReConnect durchfhren
				DapiFTDIdoReConnect(handle32);
				++DapiHandle->packet_statistic_transmit_multiple_count_recon;
			}
			else
			{
				// Empfangsbuffer flushen
				ret = FTread_n_bytes_with_timeout(handle32, 200, buff, 50*1000);	// Read x Byte (max 50ms->zum flushen)
			}

			// DEBUG
			debug_set_function_name("DapiFTDISendCommandMultipleByte");
			sprintf(debug_msg, "FLUSH BUFFER - buff_cnt_ret = %ld bytes wurden geflusht", ret);
			debug_collect(debug_msg);
			if (ret > 0)
			{
				for (i=0; i<ret; i+=16)
				{
					debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), buff, ret, i, ret);
					debug_collect(debug_msg);
				}
			}
			// -----
		}

		buff_cnt=0;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
		buff[buff_cnt++] = 0xfe;
	
		buff[buff_cnt++] = 0xff;
		buff[buff_cnt++] = (unsigned char) DapiHandle->jobID;

		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			buff[buff_cnt++] = (unsigned char) (command | 0x80);		// Kennzeichne "neues Protokoll"
		}
		else
		{
			buff[buff_cnt++] = (unsigned char) command;
		}

		buff[buff_cnt++] = (unsigned char) ((adress      ) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >>  8) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >> 16) & 0xff);
		buff[buff_cnt++] = (unsigned char) ((adress >> 24) & 0xff);

		buff[buff_cnt++] = (unsigned char) be;
		// das wird IMMER gesendet

		buff[buff_cnt++] = (unsigned char)  adress_tiefe;				// Adressraum-Tiefe
		buff[buff_cnt++] = (unsigned char) (adress_tiefe>>8) & 0xff;	// Adressraum-Tiefe
		buff[buff_cnt++] = (unsigned char)  wiederholungen;				// Anzahl Wiederholungen
		buff[buff_cnt++] = (unsigned char) (wiederholungen>>8) & 0xff;	// Anzahl Wiederholungen

		checksum_calc = 0;
		checksum_calc += (unsigned char) buff[buff_cnt - 11];
		checksum_calc += (unsigned char) buff[buff_cnt - 10];
		checksum_calc += (unsigned char) buff[buff_cnt - 9];
		checksum_calc += (unsigned char) buff[buff_cnt - 8];
		checksum_calc += (unsigned char) buff[buff_cnt - 7];
		checksum_calc += (unsigned char) buff[buff_cnt - 6];
		checksum_calc += (unsigned char) buff[buff_cnt - 5];
		checksum_calc += (unsigned char) buff[buff_cnt - 4];
		checksum_calc += (unsigned char) buff[buff_cnt - 3];
		checksum_calc += (unsigned char) buff[buff_cnt - 2];
		checksum_calc += (unsigned char) buff[buff_cnt - 1];

		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			buff[buff_cnt++] = (unsigned char) (checksum_calc       & 0xff);			// Checksum 
			buff[buff_cnt++] = (unsigned char) ((checksum_calc >> 8) & 0xff);			// Checksum 
		}

		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "send_char_cnt = %ld", buff_cnt);
		debug_collect(debug_msg);
		// -----
		
		if(command == DT_FTDI_USB_CMD_SENDNBYTE)
		{
			// 
			for(i=0;i!=wiederholungen;++i)
			{
				for(j=0;j!=adress_tiefe;++j)
				{
					rd_offset = i*(adress_tiefe) + j;

					if(rd_offset < data_length) 
					{
						buff[buff_cnt++] = *(data + rd_offset);
					}
					else
					{
						buff[buff_cnt++] = 0;
					}
				}
			}
		}

		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "send_char_cnt = %d", (unsigned int) buff_cnt);
		debug_collect(debug_msg);
		// -----

		ULONG wr_pos = 0;
		ULONG wr_size = 64*1024;			// Darf wg. Timeout nicht hoeher sein
		while(buff_cnt > 0)
		{
			if(buff_cnt > wr_size)
			{
				dStatus = FT_Write(DapiHandle->ftHandle, buff + wr_pos, wr_size, &buff_cnt_ret);
				
				// DEBUG
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "FT_Write( gibt zurck : buff_cnt_ret=%d", (unsigned int) buff_cnt_ret);
				debug_collect(debug_msg);
				// -----
				
				wr_pos += wr_size;
				buff_cnt -= wr_size;
			}
			else
			{
				dStatus = FT_Write(DapiHandle->ftHandle, buff + wr_pos, buff_cnt, &buff_cnt_ret);
				
				// DEBUG
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "FT_Write( gibt zurck : buff_cnt_ret=%d", (unsigned int) buff_cnt_ret);
				debug_collect(debug_msg);
				// -----
				
				buff_cnt = 0;
			}
		}
	
		if(dStatus != FT_OK)
		{
			DapiFTDIdoReConnect(handle32);
		}
	
		if(IsFTDIError(dStatus)) 
		{
			// DEBUG
			debug_set_function_name("DapiFTDISendCommandMultipleByt");
			sprintf(debug_msg, "ERROR FT_Write");
			debug_collect(debug_msg);
			debug_collect_printall();
			// -----
		
			--second_call_prevent;	// prevent a background call 
			free(buff);

			return DAPI_ERR_COM_FT_ERROR;	// Fehler
		}
		
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "B");
		debug_collect(debug_msg);
		// -----
		
		if(command == DT_FTDI_USB_CMD_GETNBYTE) 
		{
			get_char_cnt = 3 + ((adress_tiefe+1)*wiederholungen);	// Start_char + job_id + n*bytes-data + end_char
		}
		if(command == DT_FTDI_USB_CMD_SENDNBYTE) 
		{
			get_char_cnt = 3 + 2;	// Start_char + job_id + 2Bytes(wiederholungen) + end_char
		}
		if(DapiHandle->ftdi_use_protocol_new == TRUE)
		{
			get_char_cnt += 4; // 4 Zeichen bei neuem Protokoll mehr !
		}

		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "get_char_cnt = %d", (unsigned int) get_char_cnt);
		debug_collect(debug_msg);
		// -----
		
		ret = FTread_n_bytes_with_timeout(handle32, get_char_cnt, buff, (get_char_cnt*50) + 400*1000);	// max ..  + 400ms Reserve

		if(ret != get_char_cnt)
		{
			// DEBUG			
			debug_set_function_name("DapiFTDISendCommandMultipleByt");
			sprintf(debug_msg, "ft_rx received char count = %d, sollte sein = %d", (unsigned int) ret, (unsigned int) get_char_cnt);
			debug_collect(debug_msg);
			// -----
			
			++error;
		}
		else
		{
			if(error!=0)
			{
				dStatus = FT_GetQueueStatus(DapiHandle->ftHandle, &buff_cnt_ret);
				if(IsFTDIError(dStatus)) 
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ERROR FT_GetQueueStatus");
					debug_collect(debug_msg);
					debug_collect_printall();
					// -----
				
					--second_call_prevent;	// prevent a background call 
					free(buff);

					return DAPI_ERR_COM_FT_ERROR;	// Fehler
				}
			}

			if(command == DT_FTDI_USB_CMD_GETNBYTE) 
			{
				buff_cnt=0;
				for(j=0;j!=wiederholungen;++j) 
				{
					if(buff[j*(adress_tiefe+1) + 2] != (j&0xff))
					{
						// DEBUG			
						debug_set_function_name("DapiFTDISendCommandMultipleByt");
						sprintf(debug_msg, "buff[j*(adress_tiefe+1)] = %x (sollte sein =%x)", (unsigned int) buff[j*(adress_tiefe+1) + 2], (unsigned int) j&0xff);
						debug_collect(debug_msg);
						// -----
						
						++error;
					}
					else
					{
						for(i=0;i!=adress_tiefe;++i) 
						{
							if(buff_cnt<data_length) 
							{
								*(data+buff_cnt++) = buff[j*(adress_tiefe+1) + i + 1 + 2];	// 1 more (filter ID for repeats; 2 more (2 bytes header
							}
						}
					}
				}
			}

			if(buff[0] != 0x1A) 
			{
				// DEBUG			
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "ft_rx_buff[0]=%x sollte sein=%x", (unsigned int) buff[0], 0x1a);
				debug_collect(debug_msg);
				// -----
				
				++error;
			}
	
			if(buff[1] != DapiHandle->jobID) 
			{
				// DEBUG			
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "ft_rx_buff[1]=%x sollte sein=%x", (unsigned int) buff[1], (unsigned int) DapiHandle->jobID);
				debug_collect(debug_msg);
				// -----
				
				++error;
			}

			// ab 10.3.2015 - Neue Ende-Zeichen....mit Error Codes mglich
		/*	if((error == 0) && ((buff[get_char_cnt-1] & 0xf8) == 0x48))
			{
				module_error_code = buff[get_char_cnt-1];

				// DEBUG			
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "Empfangener ERROR Code vom Modul = 0x%x", module_error_code);
				debug_collect(debug_msg);
				// -----
			}
			else
*/
			if(DapiHandle->ftdi_use_protocol_new != TRUE)
			{
				// ALTES Protokoll
				if(buff[get_char_cnt-1] != 0x4f)
				{
					// DEBUG			
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_buff[%d]=%x sollte sein=0x4f", (unsigned int) get_char_cnt, (unsigned int) buff[get_char_cnt-1]);
					debug_collect(debug_msg);
					// -----
					
					++error;
				}
			}
			else
			{
				// NEUES Protokoll
				// Check Zeichen: Ende altes Protokoll
				if(buff[get_char_cnt-5] != 0x4f)
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_buff[%d]=%x sollte sein=0x%x", (unsigned int) get_char_cnt-4, (unsigned int) buff[get_char_cnt-4], 0x4f);
					debug_collect(debug_msg);
					// -----
					
					++error;
				}

				// 1. Byte "neues" Protokoll // ERROR Code
				module_error_code = (buff[get_char_cnt-4] & 0xff);
				if(module_error_code != DAPI_ERR_NONE)
				{
					module_error_code |= DAPI_ERR_DEV_CLASS;
				}		

				// DEBUG			
				debug_set_function_name("DapiFTDISendCommandMultipleByt");
				sprintf(debug_msg, "Empfangener ERROR Code vom Modul = 0x%lx", module_error_code);
				debug_collect(debug_msg);
				// -----

				switch(module_error_code)
				{
					case DAPI_ERR_NONE:						break;
					case DAPI_ERR_DEV_IO_ADDR_ILLEGAL:		break;
					case DAPI_ERR_DEV_IO_OTHER_ERROR:		break;
					case DAPI_ERR_DEV_IO_RD_WR_ERROR:		break;
					case DAPI_ERR_DEV_IO_SUB_ADDR_ILLEGAL:	break;
					default:	++error;					break;
				}

				// Check 2.+3. Byte "neues" Protokoll / CHECKSUM
				checksum_rx  =  (unsigned long) (buff[get_char_cnt-3] & 0xff);
				checksum_rx += ((unsigned long) (buff[get_char_cnt-2] & 0xff) << 8);
			/*	if(checksum_rx != (checksum_calc & 0xffff))
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_ checksum_rx=%x sollte sein=0x%x (CHECKSUM ERROR)", checksum_rx, (unsigned int) (checksum_calc & 0xffff));
					debug_collect(debug_msg);
					// -----
					
					++error;
				}*/


				// Check 4. Byte "neues" Protokoll
				if(buff[get_char_cnt-1] != 0xaf)
				{
					// DEBUG
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_buff[%d]=%x sollte sein=0x%x", (unsigned int) get_char_cnt-1, (unsigned int) buff[get_char_cnt-1], 0xaf);
					debug_collect(debug_msg);
					// -----
					
					++error;
				}
			}

			// ------------------------
			if(command == DT_FTDI_USB_CMD_SENDNBYTE) 
			{
				if(buff[2] != (wiederholungen&0xff)) 
				{
					// DEBUG			
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_buff[2]=%x sollte sein=%x", (unsigned int) buff[2], (unsigned int) (wiederholungen&0xff));
					debug_collect(debug_msg);
					// -----
					
					++error;
				}
				if(buff[3] != ((wiederholungen>>8)&0xff)) 
				{
					// DEBUG			
					debug_set_function_name("DapiFTDISendCommandMultipleByt");
					sprintf(debug_msg, "ft_rx_buff[3]=%x sollte sein=%x", (unsigned int) buff[3], (unsigned int) ((wiederholungen>>8)&0xff));
					debug_collect(debug_msg);
					// -----
					
					++error;
				}
			}
			// ------------------------
		}

		if(error!=0) 
		{
			// DEBUG			
			debug_set_function_name("DapiFTDISendCommandMultipleByt");
			sprintf(debug_msg, "ft_rx_buff[%d]=%x sollte sein=%x", (unsigned int) get_char_cnt, (unsigned int) buff[get_char_cnt-1], (unsigned int) 0x4f);
			debug_collect(debug_msg);
			sprintf(debug_msg,"DapiFTDISendCommandMultipleBytes - RETRY RETRY RETRY RETRY = %ld / %ld", (unsigned long) retry_cnt, (unsigned long) 2*DapiHandle->retry);
			debug_collect(debug_msg);
			// -----
		}

		if(error != 0)
		{
			++retry_cnt;
		}

	} while((error!=0) && (retry_cnt <= (2*DapiHandle->retry)));

	--second_call_prevent;	// prevent a background call 
	free(buff);

	if(error==0) 
	{	
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");
		sprintf(debug_msg, "DapiFTDISendCommandMultipleByt END * module_error_code = 0x%lx* return = 0x%lx-", module_error_code, ret);
		debug_collect(debug_msg);
		// -----

		if(module_error_code != DAPI_ERR_NONE)
		{
			debug_collect_printall();
		}
		else if(retry_cnt > 0)
		{
			//debug_collect_printall();
		}

		return module_error_code;
	}
	else 
	{
		// DEBUG
		debug_set_function_name("DapiFTDISendCommandMultipleByt");

		sprintf(debug_msg, "DapiFTDISendCommandMultipleByt END * return = DAPI_ERR_DEV_PKT_OTHER_ERROR");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_DEV_PKT_OTHER_ERROR;	// Fehler
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIPing(ULONG handle32, ULONG value)
{
	unsigned long delib_error_code;
	unsigned char buff[8];
	ULONG data;

	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_PING, value, 0, 0, 0, (unsigned char*) &buff, 4);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return -1;
	}

	data=0;
	data |= (buff[0] & 0xff);
	data |= (buff[1] & 0xff) << 8;
	data |= (buff[2] & 0xff) << 16;
	data |= (buff[3] & 0xff) << 24;
	return data;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiFTDIWriteByte(ULONG handle32, ULONG adress, ULONG value)
{
	unsigned long delib_error_code;
	ULONG be;
	ULONG i;
	unsigned char buff[8];

	for(i=0;i!=8;++i) buff[i]=0;

	buff[adress&7] = (unsigned char) (value&0xff);

	be = 1 << (adress&7);
	
	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_SEND8BYTE, adress, be, (unsigned char*) &buff, 8, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	return;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiFTDIWriteWord(ULONG handle, ULONG adress, ULONG value)
{
	unsigned long delib_error_code;
	ULONG be;
	ULONG i;
	unsigned char buff[8];

	for(i=0;i!=8;++i) buff[i]=0;

	buff[(adress&6)+0] = (unsigned char) ( value     & 0xff);
	buff[(adress&6)+1] = (unsigned char) ((value>>8) & 0xff);

	be = 3 << (adress&6);
	
	PacketStatisticStart(handle);
	delib_error_code = DapiFTDISendCommand(handle, DT_FTDI_USB_CMD_SEND8BYTE, adress, be, (unsigned char*) &buff, 8, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	return;

}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiFTDIWriteLong(ULONG handle32, ULONG adress, ULONG value)
{
	unsigned long delib_error_code;
	ULONG be;
	ULONG i;
	unsigned char buff[8];

	for(i=0;i!=8;++i) buff[i]=0;

	buff[(adress&4)+0] = (unsigned char) ( value     & 0xff);
	buff[(adress&4)+1] = (unsigned char) ((value>>8) & 0xff);
	buff[(adress&4)+2] = (unsigned char) ((value>>16) & 0xff);
	buff[(adress&4)+3] = (unsigned char) ((value>>24) & 0xff);

	be = 15 << (adress&4);
	
	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_SEND8BYTE, adress, be, (unsigned char*) &buff, 8, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	return;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiFTDIWriteLongLong(ULONG handle32, ULONG adress, ULONGLONG value)
{
	unsigned long delib_error_code;
	ULONG be;

	be = 255;
	
	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_SEND8BYTE, adress, be, (unsigned char*) &value, 8, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	return;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIReadByte(ULONG handle32, ULONG adress)
{
	unsigned long delib_error_code;
	ULONG data;
	ULONG be;
	unsigned char buff[8];

	be = 1 << (adress&7);

	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_GET8BYTE, adress, be, 0, 0, (unsigned char*) &buff, 8);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	else
	{
		data=0;
		data |= (buff[adress&7] & 0xff);
	}
	return data;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIReadWord(ULONG handle32, ULONG adress)
{
	unsigned long delib_error_code;
	ULONG data;
	ULONG be;
	unsigned char buff[8];

	be = 3 << (adress&6);

	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_GET8BYTE, adress, be, 0, 0, (unsigned char*) &buff, 8);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	else
	{
		data=0;
		data |= (buff[(adress&6)+0] & 0xff);
		data |= (buff[(adress&6)+1] & 0xff) << 8;
	}
	return data;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIReadLong(ULONG handle32, ULONG adress)
{
	unsigned long delib_error_code;
	ULONG data;
	ULONG be;
	unsigned char buff[8];

	be = 15 << (adress&4);

	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_GET8BYTE, adress, be, 0, 0, (unsigned char*) &buff, 8);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	else
	{
		data=0;
		data |= (buff[(adress&4)+0]&0xff);
		data |= (buff[(adress&4)+1]&0xff) << 8;
		data |= (buff[(adress&4)+2]&0xff) << 16;
		data |= (buff[(adress&4)+3]&0xff) << 24;
	}
	return data;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONGLONG DapiFTDIReadLongLong(ULONG handle32, ULONG adress)
{
	unsigned long delib_error_code;
	ULONGLONG data;
	ULONG be;

	be = 255;

	PacketStatisticStart(handle32);
	delib_error_code = DapiFTDISendCommand(handle32, DT_FTDI_USB_CMD_GET8BYTE, adress, be, 0, 0, (unsigned char*) &data, 8);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}

	return data;
}
		
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIReadMultipleBytes(ULONG handle, ULONG adress, ULONG adress_depth, ULONG increments, UCHAR * buff, ULONG buff_len)
{
	unsigned long delib_error_code;
	ULONG pos=0;
	ULONG incr=increments;
	ULONG send_incr;
	ULONG buff_offset = 0;

	PacketStatisticStart(handle);

	while(incr>0)
	{
		if(incr>32768)
		{
			send_incr = 32768;
		}
		else
		{
			send_incr = incr;
		}
		
		delib_error_code = DapiFTDISendCommandMultipleBytes(handle, DT_FTDI_USB_CMD_GETNBYTE, adress, adress_depth, send_incr, buff + buff_offset, buff_len-buff_offset);
		if(delib_error_code != DAPI_ERR_NONE)
		{
			PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);
			DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
			return 0;
		}

		incr -= send_incr;
		buff_offset += send_incr*adress_depth;
	}
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiFTDIWriteMultipleBytes(ULONG handle, ULONG adress, ULONG adress_depth, ULONG increments, UCHAR * buff, ULONG buff_len)
{
	unsigned long delib_error_code;
	ULONG pos=0;
	ULONG incr=increments;
	ULONG send_incr;
	ULONG buff_offset = 0;

	PacketStatisticStart(handle);

	while(incr>0)
	{
		if(incr>32768)
		{
			send_incr = 32768;
		}
		else
		{
			send_incr = incr;
		}
		
		delib_error_code = DapiFTDISendCommandMultipleBytes(handle, DT_FTDI_USB_CMD_SENDNBYTE, adress, adress_depth, send_incr, buff + buff_offset, buff_len-buff_offset);
		if(delib_error_code != DAPI_ERR_NONE)
		{
			DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
			PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);
			return 0;
		}

		incr -= send_incr;
		buff_offset += send_incr*adress_depth;
	}
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
