//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	deditec_debug_comfort.c
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	26.11.2013	Start DEDITEC-History
// CC	17.12.2013	Added function debug_sprintf and debug_strcat



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// debug_print
// debug_info
// debug_error
// ------------------------------------
// debug_sprintf
// debug_strcat
// ------------------------------------
// debug_print_hex
// debug_strcat_hex

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#ifdef _MSC_VER
	#include <windows.h>	// OutputDebugString
#endif

#include <stdarg.h>		// va_list(), va_start(), va_end()
#include <stdio.h>		// vsprintf()

#ifdef DELIB_COMPILE
	#include "../../proj_config.h"
#endif

#include "../vc/deditec_string_utils.h"

// ------------------------------------
// ------------------------------------
// Global Vars
// ------------------------------------
// ------------------------------------

#ifdef DELIB_COMPILE
	char debug_msg_2k[2000];
#else
	char debug_msg_2k[200];
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifndef __linux__

void debug_print(const char * format, ...)
{
	#ifdef DEBUG_CODE_FUER_DEBUG_PRINT_INFO_ERROR_SOLL_MIT_COMPILIERT_WERDEN
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
		#else
			vsprintf(debug_msg_2k, format, args);
		#endif

		va_end(args);

		debug_out_print(debug_msg_2k);
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_info(const char * format, ...)
{
	va_list args;

	va_start(args, format);
	
	#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
		vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
	#else
		vsprintf(debug_msg_2k, format, args);
	#endif

	va_end(args);

	debug_out_info(debug_msg_2k);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_error(const char * format, ...)
{
	va_list args;

	debug_out_error("ERROR * ERROR * ERROR ERROR * ERROR * ERROR ");

	va_start(args, format);	

	#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
		vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
	#else
		vsprintf(debug_msg_2k, format, args);
	#endif

	va_end(args);

	debug_out_error(debug_msg_2k);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_sprintf(char *dest_buffer, const char *format, ...)
{
	#ifdef DEBUG_CODE_FUER_DEBUG_PRINT_INFO_ERROR_SOLL_MIT_COMPILIERT_WERDEN
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
		#else
			vsprintf(debug_msg_2k, format, args);
		#endif

		va_end(args);
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_strcat(char *dest_buffer, const char *format, ...)
{
	#ifdef DEBUG_CODE_FUER_DEBUG_PRINT_INFO_ERROR_SOLL_MIT_COMPILIERT_WERDEN
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
		#else
			vsprintf(debug_msg_2k, format, args);
		#endif

		va_end(args);
		
		strcat_n(dest_buffer, sizeof(dest_buffer), debug_msg_2k, strlen_n(debug_msg_2k, sizeof(debug_msg_2k)));
	#else
		return;
	#endif
}

#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void sprintf_append(char *dest_buffer, unsigned long dest_size, const char *format, ...)
{
	va_list args;
	va_start(args, format);	

	#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
		vsprintf_s(debug_msg_2k, sizeof(debug_msg_2k), format, args);
	#else
		vsprintf(debug_msg_2k, format, args);
	#endif

	va_end(args);

	strcat_n(dest_buffer, dest_size, debug_msg_2k, strlen_n(debug_msg_2k, sizeof(debug_msg_2k)));
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_hexdump_line_to_msg(char *msg, unsigned long msg_size, unsigned char *buffer, unsigned long buffer_size, unsigned long pos_start, unsigned long pos_max)
{
	unsigned long i;
	
	msg[0] = 0;	// Clear String
	sprintf_append(msg, msg_size, "%04lX: ", pos_start);

	for (i = 0; i != 16; i++)
	{
		if (((i + pos_start) < buffer_size) && ((i + pos_start) < pos_max))
		{
			sprintf_append(msg, msg_size, "%02X ", buffer[i+pos_start]);
		}
		else
		{
			sprintf_append(msg, msg_size, "   ");	// 3 leerzeichen
		}
		if (i == 7) sprintf_append(msg, msg_size, " ");
	}
	
	sprintf_append(msg, msg_size, "# ");

	for (i = 0; i != 16; i++)
	{
		if (((i + pos_start) < buffer_size) && ((i + pos_start) < pos_max))
		{
			if ((buffer[i+pos_start] > 0x20) && (buffer[i+pos_start] < 0x7f) && (buffer[i+pos_start] != '%'))
			{
				sprintf_append(msg, msg_size, "%c", buffer[i+pos_start]);
			}
			else
			{
				sprintf_append(msg, msg_size, ".");
			}
			if (i == 7) sprintf_append(msg, msg_size, " ");
		}
	}
	// das macht besser "der aufrufer" - debug_info, etc printen immer "\n\r" am ende dazu.. dann wre das doppelt!
	//sprintf_append(msg, msg_size, "\n\r");
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
