//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	deditec_tcp_io.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2017
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT	1
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// tcp_client_connect
// get_eth_mac
// tcp_disconnect

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>
#include <stdlib.h>        	// exit()
#include <errno.h>
#include <signal.h>        	// struct sigaction, SIGALARM
#include <unistd.h> 		// read(), close()

#include "deditec_tcp_io.h"
#include "deditec_string_utils.h"
#include "deditec_buff_utils.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	#include <strings.h>    // recv()
	#include <sys/socket.h> // socket(), send(), connect()
	#include <sys/types.h>  // socket(), send(), connect()
	#include <netinet/in.h> // htons(), defines struct sockaddr_in
	#include <netdb.h>
	#include <sys/ioctl.h>
	#include <net/if.h>
	#include <fcntl.h>
	#include <string.h>		// memcpy()
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include <winsock.h>    // socket(), send(), recv(), connect()
#endif

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

#ifndef __linux__
extern void OutputDebugString_x64(char * msg);
#endif

#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
	#define	debug_print(msg)	{OutputDebugString_x64(msg);}
#else
	#define	debug_print(msg)	OutputDebugString(msg)
#endif


#ifdef __linux__
	#ifdef DEBUG_OUTPUT
		#define	debug_print(msg)	printf("%s\n\r", msg);
	#endif
#endif


#ifndef __linux__
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	// LINKER: Add wsock32.lib
   #pragma comment(lib, "wsock32.lib ")
#endif

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// debug_print(debug_debug_msg)										nur bei DEBUG_AUSGABE		nur bei DEBUG_AUSGABE

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_client_connect(TCP_IO        * tcp_params,
                                 unsigned char * host_name, unsigned long host_name_size,
                                 unsigned long   port_no,   unsigned long keep_alive,   unsigned long tcp_timeout)
{
	struct sockaddr_in serv_addr;
	struct timeval time_tcp_timeout;
	fd_set set;

	struct hostent *server;

	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
	#endif

	unsigned char host_name_cp[IP_ADDR_SERVER_SIZE];

	// -------------------------------------- init ----------------------------------
	#ifndef __linux__
		WSADATA wsaData;

		// DEBUG
		#ifdef DEBUG_OUTPUT
			sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|   tcp_client_connect START");
			debug_print(debug_msg);
		#endif
		// -----

		if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
		{
			// DEBUG
			#ifdef DEBUG_OUTPUT
				sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|return is %d dec, loading windows socket (winsock) failed", -1);
				debug_print(debug_msg);
			#endif
			// -----
			
			return (-1);
		}
	#endif

	tcp_params->tcp_handle_tx = 0;

	strcpy_n((char*) host_name_cp, IP_ADDR_SERVER_SIZE, (char*) host_name, host_name_size ); 

	// -------------------------------------- begin ----------------------------------
	if (tcp_params->tcp_handle_tx == 0) 
	{
		// DEBUG
		#ifdef DEBUG_OUTPUT
			sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|if (tcp_params->tcp_handle_tx == 0) **** wird ausgefhrt");
			debug_print(debug_msg);
		#endif

		tcp_params->tcp_handle_tx = (unsigned long) socket (AF_INET, SOCK_STREAM, 0); 

		if (tcp_params->tcp_handle_tx < 0)
		{

			// DEBUG
			#ifdef DEBUG_OUTPUT
				sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|return is %d dec, creating linux-socket failed", -2);
				debug_print(debug_msg);
			#endif
			// -----
			
			return -2;
		}

		memset_n((unsigned char*) &serv_addr.sin_zero, sizeof(serv_addr.sin_zero), 0, sizeof(serv_addr.sin_zero)); 
		serv_addr.sin_family = AF_INET;

		serv_addr.sin_port = (unsigned short) htons((unsigned short) port_no); 

		server = gethostbyname((const char *) host_name_cp);
		if (server == NULL)
		{
			// DEBUG
			#ifdef DEBUG_OUTPUT
				sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|return is %d dec, could not resolve host name", -3);
				debug_print(debug_msg);
			#endif
			// -----

			tcp_disconnect(tcp_params);

			return (-3); 
		}

		memcpy(&(serv_addr.sin_addr),server->h_addr,server->h_length);
	}

	time_tcp_timeout.tv_sec = (tcp_timeout / 1000);
	time_tcp_timeout.tv_usec = (tcp_timeout % 1000) * 1000;
	
	FD_ZERO(&set);
	FD_SET(tcp_params->tcp_handle_tx, &set);

	unsigned long ulNonBlockingMode = 1;
	#ifdef __linux__
		ioctl(tcp_params->tcp_handle_tx, FIONBIO, &ulNonBlockingMode); 
	#else
		ioctlsocket(tcp_params->tcp_handle_tx, FIONBIO, &ulNonBlockingMode);
	#endif

	// DEBUG
	#ifdef DEBUG_OUTPUT
		sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|Connect");
		debug_print(debug_msg);
	#endif
	// -----

	if (connect(tcp_params->tcp_handle_tx, (const struct sockaddr *) &serv_addr, sizeof(serv_addr)) == -1)
	{
		#ifdef __linux__
			if( errno != EINPROGRESS )
		#else
			if ((WSAGetLastError() != WSAEINPROGRESS) && (WSAGetLastError() != WSAEWOULDBLOCK))
		#endif
		{

			// DEBUG
			#ifdef DEBUG_OUTPUT
				#ifdef __linux__
					sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|return is %d dec, failed connecting to server*errno=%d", -4, errno);
				#else
					sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|return is %d dec, failed connecting to server*errno=%d", -4, WSAGetLastError());
				#endif
				debug_print(debug_msg);
			#endif
			// -----

			tcp_disconnect(tcp_params);

			return (-4);
		}
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|Select - A");
		debug_print(debug_msg);
	#endif
	// -----

	int status = select(tcp_params->tcp_handle_tx + 1, NULL, &set, NULL, &time_tcp_timeout);
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|Select - B");
		debug_print(debug_msg);
	#endif
	// -----

	if(status == 0)
	{
		// DEBUG
		#ifdef DEBUG_OUTPUT
			sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|failed connecting to server*function select - status is %d dec", status);
			debug_print(debug_msg);
		#endif
		// -----
			
		tcp_disconnect(tcp_params);

		return (-4);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		sprintf(debug_msg,"|DEDITEC deditec_tcp_io.cpp|END - ALL OK");
		debug_print(debug_msg);
	#endif
	// -----


	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifdef __linux__
	unsigned long get_eth_mac( unsigned char * buffer, unsigned long buffer_length, unsigned long eth_no)
	{
		int s;
		struct ifreq buffer_mac;
		unsigned char eth_string[6];
		unsigned char eth_no_string[3];
		unsigned long mac_addr_found;

		if (buffer_length < 6 || eth_no > 99) return -1;

		s = socket(PF_INET, SOCK_DGRAM, 0);
		
		if (s == -1) return -1;

		strcpy_n(eth_string, sizeof(eth_string), "eth", 3);
		snprintf(eth_no_string, sizeof(eth_no_string), "%d", (int) eth_no);
		strcpy_n(eth_string + 3, sizeof(eth_string -3), eth_no_string, 3);

		strcpy_n(buffer_mac.ifr_name, sizeof(buffer_mac.ifr_name), eth_string, strlen_n(eth_string, sizeof(eth_string)));

		ioctl(s, SIOCGIFHWADDR, &buffer_mac);

		close(s);

		mac_addr_found = 0;

		for( s = 0; s < 6; s++ )
		{
			*(buffer + s) = *(buffer_mac.ifr_hwaddr.sa_data +s);

			if ( *(buffer + s) != 0) mac_addr_found = 1;
		}

		if (mac_addr_found == 1) return 0;
		else return -1;
	}
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_disconnect(TCP_IO * tcp_params)
{
	char debug_msg[200];
	unsigned long ret;
	ret = 0;

	if (tcp_params->tcp_handle_tx > 0)
	{
		#ifdef __linux__
			ret = close(tcp_params->tcp_handle_tx);
		#else
			ret = closesocket(tcp_params->tcp_handle_tx);

			sprintf(debug_msg,"|tcp_disconnect( - tcp_params->tcp_handle_tx < 0 - closesocket returns 0x%x", ret);
			debug_print(debug_msg);

			WSACleanup(); 
		#endif

		if (ret == 0)
		{
			tcp_params->tcp_handle_tx = 0;
		}
	}
	else
	{
		#ifdef DEBUG_OUTPUT
		sprintf(debug_msg,"|tcp_disconnect( - tcp_params->tcp_handle_tx < 0 - closesocket NOT POSSIBLE");
		debug_print(debug_msg);
		#endif
	}
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
