//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_Dapi_di_do_ad_da_befehle.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
// 
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiDOSet1
// DapiDOSet8
// DapiDOSet16
// DapiDOSet32
// DapiDOSet64
// DapiDOReadback32
// DapiDOReadback64
// DapiDOSet1_WithTimer
// DapiDOSetBit32
// DapiDOClrBit32
// ------------------------------------
// DapiDIGet1
// DapiDIGet8
// DapiDIGet16
// DapiDIGet32
// DapiDIGet64
// DapiDIGetFF32
// DapiDIGetCounter
// ------------------------------------
// DapiADGet
// DapiADGetValueModeUnit
// DapiADGetVolt
// DapiADGetmA
// DapiADSetMode
// DapiADGetMode
// ------------------------------------
// DapiDASet
// DapiDASetVolt
// DapiDASetmA
// DapiDAGetValueModeUnit
// DapiDASetMode
// DapiDAGetMode
// ------------------------------------
// DapiTempGet
// ------------------------------------
// DapiCnt48ModeSet
// DapiCnt48ModeGet
// DapiCnt48CounterGet32
// DapiCnt48CounterGet48
// ------------------------------------
// DapiPulseGenSet
// ------------------------------------
// DapiPWMOutSet
// DapiPWMOutReadback
// ------------------------------------
// DapiStepperCommand
// DapiStepperCommandEx
// DapiStepperGetStatus
// ------------------------------------
// DapiScanModule
// ------------------------------------
// DapiWatchdogEnable
// DapiWatchdogDisable
// DapiWatchdogRetrigger

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	#ifdef USE_FTDI
		#include "../usb-ftdi/ftd2xx.h"
	#endif

	#include "library/vc/deditec_tcp_io.h"	// muss vor die "delib_intern.h", weil hier TCP_IO drin ist
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
	#include "library/communication/dt_communication_paramams.h"
	
	#include "delib_hdlmgr.h"
	#include "delib_set_get_error.h"
	#include "delib_open_close.h"
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include <windows.h>
	#include "conio.h"
	#include "delib_io_tcp.h"
	#include "delib_Dapi_prog_device_functions.h"
	#include "delib_io_ftdi.h"
#endif

// ------------------------------------
// ------------------------------------
// Global Vars
// ------------------------------------
// ------------------------------------

unsigned char global_ad_buffer[128*4];		// Max 128 chan

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet1_WithTimer(ULONG handle32, ULONG ch, ULONG data, ULONG time_ms)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOSet1_WithTimer(handle32=0x%lx, ch=%ld, data=0x%lx, time_ms=%ld (dez.))", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data, (unsigned long) time_ms);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet1_WithTimer");
		sprintf(debug_msg, "DapiDOSet1_WithTimer(handle32=0x%lx, ch=%ld, data=0x%lx, time_ms=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data, (unsigned long) time_ms);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet1_WithTimer -> Module does not support this command:DAPI_SW_FEATURE_BIT_CFG_DO");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO_WITH_TIMER) == 0)
		{
			sprintf(debug_msg, "DapiDOSet1_WithTimer -> Module does not support this command:DAPI_SW_FEATURE_BIT_CFG_DO_WITH_TIMER");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)
	{
		sprintf(debug_msg, "DapiDOSet1_WithTimer -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteLongXXX(handle32, 0x200, time_ms);
		DapiWriteWordXXX(handle32, 0x204, (ch << 8) | (data&1));
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_WITH_TIMER;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		ULONG value = (time_ms & 0x7fffffff) | ((data & 1) << 31);
		DapiWriteLongXXX(handle32, adr, value);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet1(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG xx;
	ULONG mask;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOSet1(handle32=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet1");
		sprintf(debug_msg, "DapiDOSet1(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)
	{
		sprintf(debug_msg, "DapiDOSet1 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet1 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
		
		// module supports DapiDOSetBit32 command ?
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO_CMD_SET_CLR_BIT_32) != 0)
		{
			// check set or clear bit
			if (data != 0)
			{
				DapiDOSetBit32XXX(handle32, (ch & (~31)), (1 << (ch & 0x1f)));
			}
			else
			{
				DapiDOClrBit32XXX(handle32, (ch & (~31)), (1 << (ch & 0x1f)));
			}
			DAPI_return;
		}
	}
	
	// old way for DapiDOSet1 (for modules with old firmware!)
	xx = DapiDOReadback32XXX(handle32, ch&0xffe0);
	mask = 1 << (ch&31);

	if(data == 0)
	{
		// clear Bit
		xx = xx & (~mask);
	}
	else
	{
		// set Bit
		xx = xx | mask;
	}
	
	DapiDOSet32(handle32, ch&0xffe0, xx);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet8(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiDOSet8(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet8");
		sprintf(debug_msg, "DapiDOSet8(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet8 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)
	{
		sprintf(debug_msg, "DapiDOSet8 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x07) != 0)	
	{
		sprintf(debug_msg, "DapiDOSet8 -> Illegal channel parameter for this command (Valid: 0, 8, 16, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteByteXXX(handle32, (ch>>3) & 0x1f, data);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		DapiWriteLongXXX(handle32, adr, data);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet16(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOSet16(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet16");
		sprintf(debug_msg, "DapiDOSet16(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet16 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)		
	{
		sprintf(debug_msg, "DapiDOSet16 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x0f) != 0)	
	{
		sprintf(debug_msg, "DapiDOSet16 -> Illegal channel parameter for this command (Valid: 0, 16, 32, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteWordXXX(handle32, (ch>>3) & (~1) & 0x1f, data);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;
		DapiWriteLongXXX(handle32, adr, data);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet32(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiDOSet32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet32");
		sprintf(debug_msg, "DapiDOSet32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet32 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)		
	{
		sprintf(debug_msg, "DapiDOSet32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDOSet32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteLongXXX(handle32, (ch>>3) & (~3) & 0x1f, data);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		DapiWriteLongXXX(handle32, adr, data);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSet64(ULONG handle32, ULONG ch, ULONGLONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOSet64(handle=0x%lx, ch=%ld, data=0x%llx)", (unsigned long) handle32, (unsigned long) ch, data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSet64");
		#ifdef __linux__
			sprintf(debug_msg, "DapiDOSet64(handle=0x%lx, ch=%ld, data=0x%llx)", (unsigned long) handle32, (unsigned long) ch, data);
		#else
			sprintf(debug_msg, "DapiDOSet64(handle=0x%lx, ch=%ld, data=0x%I64x)", (unsigned long) handle32, (unsigned long) ch, data);
		#endif
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOSet64 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)		
	{
		sprintf(debug_msg, "DapiDOSet64 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x3f) != 0)	
	{
		sprintf(debug_msg, "DapiDOSet64 -> Illegal channel parameter for this command (Valid: 0 or 64)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteLongLongXXX(handle32, (ch>>3) & (~7) & 0x1f, data);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONGLONG;
		DapiWriteLongLongXXX(handle32, adr, data);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSetBit32(ULONG handle32, ULONG ch, ULONG data)
{
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOSetBit32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	DapiDOSetBit32XXX(handle32, ch, data);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOSetBit32XXX(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG addr;
	addr = 0x80 | ((ch>>3) & 0x1f);
	ULONG adr;
	ULONGLONG handle_ull;
	
	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOSetBit32");
		sprintf(debug_msg, "DapiDOSetBit32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module does not support new Software-Features OR
	// module does not have the accessed I/O-type OR 
	// command is not supported ?
	if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
		 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0) || \
		 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO_CMD_SET_CLR_BIT_32) == 0) )
	{
		sprintf(debug_msg, "DapiDOSetBit32 -> Module does not support this command");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
		return;
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)
	{
		sprintf(debug_msg, "DapiDOSetBit32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDOSetBit32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteLongXXX(handle32, addr, data);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_SET_BIT;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		DapiWriteLongXXX(handle32, addr, data);
	}	
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOClrBit32(ULONG handle32, ULONG ch, ULONG data)
{
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOClrBit32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	DapiDOClrBit32XXX(handle32, ch, data);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDOClrBit32XXX(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG addr;
	ULONG adr;
	
	addr = 0xa0 | ((ch>>3) & 0x1f);
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOClrBit32");
		sprintf(debug_msg, "DapiDOClrBit32(handle=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module does not support new Software-Features OR
	// module does not have the accessed I/O-type OR 
	// command is not supported ?
	if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
		 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0) || \
		 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO_CMD_SET_CLR_BIT_32) == 0) )
	{
		sprintf(debug_msg, "DapiDOClrBit32 -> Module does not support this command");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
		return;
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)
	{
		sprintf(debug_msg, "DapiDOClrBit32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDOClrBit32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteLongXXX(handle32, addr, data);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_CLEAR_BIT;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		DapiWriteLongXXX(handle32, addr, data);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGet1(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	unsigned char xx;
	unsigned char mask;
	ULONG ret;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGet1(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
		{
			sprintf(debug_msg, "DapiDIGet1 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGet1 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	xx = (unsigned char) DapiDIGet8(handle32, ch&0xfff8);
	mask = 1 << (ch&7);
	
	if((xx&mask) == 0)
	{
		ret = 0;
	}
	else
	{
		ret = 1;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGet1");
		sprintf(debug_msg, "DapiDIGet1(handle=0x%lx, ch=%ld) * ret=%ld", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGet8(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGet8(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
		{
			sprintf(debug_msg, "DapiDIGet8 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGet8 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x07) != 0)	
	{
		sprintf(debug_msg, "DapiDIGet8 -> Illegal channel parameter for this command (Valid: 0, 8, 16, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadByteXXX(handle32, 0x20 | ((ch>>3) & 0x1f));
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DI;
		adr |= DEV_A23_16_GROUP_DI_SUB_DI_RD_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGet8");
		sprintf(debug_msg, "DapiDIGet8(handle=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGet16(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGet16(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
		{
			sprintf(debug_msg, "DapiDIGet16 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGet16 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x0f) != 0)	
	{
		sprintf(debug_msg, "DapiDIGet16 -> Illegal channel parameter for this command (Valid: 0, 16, 32, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadWordXXX(handle32, 0x20 | ((ch>>3) & 0x1f));
	}
	else
	{
		adr = DEV_A31_24_GROUP_DI;
		adr |= DEV_A23_16_GROUP_DI_SUB_DI_RD_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGet16");
		sprintf(debug_msg, "DapiDIGet16(handle=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGet32(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGet32(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
		{
			sprintf(debug_msg, "DapiDIGet32 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGet32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDIGet32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongXXX(handle32, 0x20 | ((ch>>3) & 0x1f));
	}
	else
	{
		adr = DEV_A31_24_GROUP_DI;
		adr |= DEV_A23_16_GROUP_DI_SUB_DI_RD_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGet32");
		sprintf(debug_msg, "DapiDIGet32(handle=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONGLONG DAPI_FUNCTION_PRE DapiDIGet64(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONGLONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGet64(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
		{
			sprintf(debug_msg, "DapiDIGet64 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGet64 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x3f) != 0)	
	{
		sprintf(debug_msg, "DapiDIGet64 -> Illegal channel parameter for this command (Valid: 0 or 64)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongLongXXX(handle32, 0x20 | ((ch>>3) & 0x1f));
	}
	else
	{
		adr = DEV_A31_24_GROUP_DI;
		adr |= DEV_A23_16_GROUP_DI_SUB_DI_RD_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONGLONG;
		ret = DapiReadLongLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGet64");
		#ifdef __linux__
			sprintf(debug_msg, "DapiDIGet64(handle=0x%lx, ch=%ld) * ret=0x%llx", (unsigned long) handle32, (unsigned long) ch, ret);
		#else
			sprintf(debug_msg, "DapiDIGet64(handle=0x%lx, ch=%ld) * ret=0x%I64x", (unsigned long) handle32, (unsigned long) ch, ret);
		#endif
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGetFF32(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGetFF32(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI_FF) == 0)
		{
			sprintf(debug_msg, "DapiDIGetFF32 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di <= ch)		
	{
		sprintf(debug_msg, "DapiDIGetFF32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDIGetFF32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongXXX(handle32, 0x40 | ((ch>>3) & 0x1f));
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DI;
		adr |= DEV_A23_16_GROUP_DI_SUB_DI_READ_FF;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGetFF32");
		sprintf(debug_msg, "DapiDIGetFF32(handle=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDIGetCounter(ULONG handle32, ULONG ch, ULONG mode)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG adr;
	ULONG ret;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDIGetCounter(handle=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI_CNT) == 0)
		{
			sprintf(debug_msg, "DapiDIGetCounter -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_di_counter <= ch)		
	{
		sprintf(debug_msg, "DapiDIGetCounter -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_di_counter);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode != TRUE)
	{
		// 16 Bit ADR-Mode
		adr = 0x100 + (ch&0x7f)*2;
		if(mode == DAPI_CNT_MODE_READ_WITH_RESET)
		{
			adr+=0x100;
		}
		else if(mode == DAPI_CNT_MODE_READ_LATCHED)
		{
			adr+=0x200;
		}
	
		ret = DapiReadWordXXX(handle32, adr);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DI;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;

		if(mode == DAPI_CNT_MODE_READ_WITH_RESET)
		{
			adr|=DEV_A23_16_GROUP_DI_SUB_CNT_READ_WITH_CLR_16;
		}
		else if(mode == DAPI_CNT_MODE_READ_LATCHED)
		{
			adr|=DEV_A23_16_GROUP_DI_SUB_CNT_READ_LATCH_16;
		}
		else
		{
			adr|=DEV_A23_16_GROUP_DI_SUB_CNT_READ_16;
		}

		ret = DapiReadLongXXX(handle32, adr);
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDIGetCounter");
		sprintf(debug_msg, "DapiDIGetCounter(handle=0x%lx, ch=%ld, mode=0x%lx) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDOReadback32(ULONG handle32, ULONG ch)
{
	char dapi_caller_text[200];
	unsigned long ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOReadback32(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ret = DapiDOReadback32XXX(handle32, ch);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDOReadback32XXX(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	ULONG adr;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOReadback32 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			return 0;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)		
	{
		sprintf(debug_msg, "DapiDOReadback32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return 0;
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x1f) != 0)	
	{
		sprintf(debug_msg, "DapiDOReadback32 -> Illegal channel parameter for this command (Valid: 0, 32, 64, ..)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongXXX(handle32, (ch>>3) & 0x1f);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		ret = DapiReadLongXXX(handle32, adr);
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOReadback32");
		sprintf(debug_msg, "DapiDOReadback32(handle=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONGLONG DAPI_FUNCTION_PRE DapiDOReadback64(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONGLONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDOReadback64(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DO) == 0)
		{
			sprintf(debug_msg, "DapiDOReadback64 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_do <= ch)		
	{
		sprintf(debug_msg, "DapiDOReadback64 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_do);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	// illegal channel parameter ? 
	if ((ch & 0x3f) != 0)	
	{
		sprintf(debug_msg, "DapiDOReadback64 -> Illegal channel parameter for this command (Valid: 0 or 64)");
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongLongXXX(handle32, (ch>>3) & 0x1f);
	}
	else
	{
		adr = DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_DO_RW_DATA8_TO_64;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONGLONG;
		ret = DapiReadLongLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDOReadback64");
		#ifdef __linux__
			sprintf(debug_msg, "DapiDOReadback64(handle=0x%lx, ch=%ld) * ret=0x%llx", (unsigned long) handle32, (unsigned long) ch, ret);
		#else
			sprintf(debug_msg, "DapiDOReadback64(handle=0x%lx, ch=%ld) * ret=0x%I64x", (unsigned long) handle32, (unsigned long) ch, ret);
		#endif
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiADSetMode(ULONG handle32, ULONG ch, ULONG mode)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG mode_rb;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiADSetMode(handle=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADSetMode");
		sprintf(debug_msg, "DapiADSetMode(handle=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_AD) == 0)
		{
			sprintf(debug_msg, "DapiADSetMode -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ai <= ch)
	{
		sprintf(debug_msg, "DapiADSetMode -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ai);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteByteXXX(handle32, 0x1000 + ch*4 + 3, mode);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_AD;
		adr |= DEV_A23_16_GROUP_AD_SUB_AD_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		DapiWriteLongXXX(handle32, adr, mode & 0xff);
	}
	
	// illegal mode (readback mode) ? 
	mode_rb = 0;

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		mode_rb = DapiReadByteXXX(handle32, 0x1000 + ch*4 + 3);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_AD;
		adr |= DEV_A23_16_GROUP_AD_SUB_AD_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		mode_rb = DapiReadLongXXX(handle32, adr);
	}
	
	if ((mode_rb != (mode&0x7fff)) && ((mode & DAPI_ADDA_MODE_PREVENT_DAPI_MODE_ERROR) == 0))	// Mit 0x8000 kann man die Fehlerausgabe unterdrcken 
	{
		sprintf(debug_msg, "DapiADSetMode -> Setting Mode = 0x%lx  readback from Module=0x%lx", (unsigned long) mode, (unsigned long) mode_rb);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADSetMode");
		sprintf(debug_msg, "DapiADSetMode(handle=0x%lx, ch=%ld) * mode_rb=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode_rb);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiADGetMode(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG mode;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiADGetMode(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_AD) == 0)
		{
			sprintf(debug_msg, "DapiADGetMode -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ai <= ch)		
	{
		sprintf(debug_msg, "DapiADGetMode -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ai);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		mode = DapiReadByteXXX(handle32, 0x1000 + ch*4 + 3);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_AD;
		adr |= DEV_A23_16_GROUP_AD_SUB_AD_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		mode = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADGetMode");
		sprintf(debug_msg, "DapiADGetMode(handle=0x%lx, ch=%ld) * ret=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return mode;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiADGet(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiADGet(handle=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_AD) == 0)
		{
			sprintf(debug_msg, "DapiADGet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ai <= (ch & 0x7fff))		
	{
		sprintf(debug_msg, "DapiADGet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ai);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if ((ch & 0x8000) == 0)
	{
		if(DapiHandle->is_Adr_32_mode == FALSE)
		{
			ret = DapiReadWordXXX(handle32, 0x1000 + ch*4);
		}
		else
		{
			adr =  DEV_A31_24_GROUP_AD;
			adr |= DEV_A23_16_GROUP_AD_SUB_AD_RD_DATA_AND_MODE;
			adr |= ch << 3;
			adr |= DEV_A07_00_LENGTH_LONG;
			ret = DapiReadLongXXX(handle32, adr);
		}
	}
	else
	{
		ret = ((ULONG) global_ad_buffer[((ch&127) * 4) + 0]);
		ret|= ((ULONG) global_ad_buffer[((ch&127) * 4) + 1]) << 8;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADGet");
		sprintf(debug_msg, "DapiADGet(handle32=0x%lx, ch=%ld) * ret=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiADGetVolt(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	char debug_msg[200];
	float value=0;
	ULONG ret_mode=0, ret_unit=DAPI_ADDA_UNIT_VOLT;
	char dapi_caller_text[200];
	ULONGLONG handle_ull;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	sprintf(dapi_caller_text, "DapiADGetVolt(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	value = DapiADGetValueModeUnitXXX(handle32, ch, &ret_mode, &ret_unit);
	
	if(ret_unit != DAPI_ADDA_UNIT_VOLT)
	{
		sprintf(debug_msg, "DapiADGetVolt -> Illegal A/D Mode detected (mode=0x%lx)", (unsigned long) ret_mode);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADGetVolt");
		sprintf(debug_msg, "DapiADGetVolt(handle32=0x%lx, ch=%ld,  ret_unit=%ld) * ret=%f", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret_unit, value);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return value;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiADGetmA(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	char debug_msg[200];
	float value=0;
	ULONG ret_mode=0, ret_unit=DAPI_ADDA_UNIT_MA;
	char dapi_caller_text[200];
	ULONGLONG handle_ull;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	sprintf(dapi_caller_text, "DapiADGetmA(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	value = DapiADGetValueModeUnitXXX(handle32, ch, &ret_mode, &ret_unit);
	
	if(ret_unit != DAPI_ADDA_UNIT_MA)
	{
		sprintf(debug_msg, "DapiADGetmA -> Illegal A/D Mode detected (mode=0x%lx)", (unsigned long) ret_mode);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
		DAPIx_return(0);
	}
	
	// Richtiger Mode - Alles ok !
		
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADGetmA");
		sprintf(debug_msg, "DapiADGetmA(handle32=0x%lx, ch=%ld) * ret_unit=0x%lx * ret=%f", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret_unit, value);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return value;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiADGetValueModeUnit(ULONG handle32, ULONG ch, ULONG *ret_mode, ULONG *ret_unit)
{
	char dapi_caller_text[200];
	float ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiADGetValueModeUnit(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ret = DapiADGetValueModeUnitXXX(handle32, ch, ret_mode, ret_unit);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiADGetValueModeUnitXXX(ULONG handle32, ULONG ch, ULONG *ret_mode, ULONG *ret_unit)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG data;
	ULONG mode;
	ULONG unit;
	float value=0;
	ULONG adr;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_AD) == 0)
		{
			sprintf(debug_msg, "DapiADGetValueModeUnit -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			return 0;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ai <= (ch & 0x7fff))		
	{
		sprintf(debug_msg, "DapiADGetValueModeUnit -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ai);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return 0;
	}
	
	if ((ch & 0x8000) == 0)
	{
		if(DapiHandle->is_Adr_32_mode == FALSE)
		{
			data = DapiReadLongXXX(handle32, 0x1000 + (ch & 0x7fff)*4);
		}
		else
		{
			adr = DEV_A31_24_GROUP_AD;
			adr |= DEV_A23_16_GROUP_AD_SUB_AD_RD_DATA_AND_MODE;
			adr |= ch << 3;
			adr |= DEV_A07_00_LENGTH_LONG;
			data = DapiReadLongXXX(handle32, adr);
		}
	}
	else
	{
		data = ((ULONG) global_ad_buffer[((ch&63) * 4) + 0]);
		data|= ((ULONG) global_ad_buffer[((ch&63) * 4) + 1]) << 8;
		data|= ((ULONG) global_ad_buffer[((ch&63) * 4) + 2]) << 16;
		data|= ((ULONG) global_ad_buffer[((ch&63) * 4) + 3]) << 24;
	}

	mode = (data >> 24) & 0xff;

	switch(mode)
	{
		case DAPI_ADDA_MODE_16BIT_BI_CAL_MODE:
			// +-5V
			value = (((float) (data&0xffff)) *10.0F / 65536.0F) - 5.0F;
			if((unsigned long) ret_unit != 0)
			{
				if(*ret_unit != 0)
				{
					unit = *ret_unit;
				}
				else
				{
					unit = DAPI_ADDA_UNIT_VOLT;
				}
			}
			break;

		case DAPI_ADDA_MODE_16BIT_UNIPOL_20V:
			// 0-20V
			value = (((float) (data&0xffff)) *20.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_UNIPOL_10V:
			// 0-10V
			value = (((float) (data&0xffff)) *10.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_UNIPOL_5V:
			// 0-5V
			value = (((float) (data&0xffff)) * 5.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_BIPOL_20V:
			// +-20V
			value = (((float) (data&0xffff)) * 40.0F / 65536.0F) - 20.0F;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_BIPOL_10V:
		case DAPI_ADDA_MODE_16BIT_BIPOL_10V_TESTMODE:
			// +-10V
			value = (((float) (data&0xffff)) * 20.0F / 65536.0F) - 10.0F;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;


			// 18 Bit !!!!!
			// 18 Bit !!!!!
			// 18 Bit !!!!!
		case DAPI_ADDA_MODE_18BIT_UNIPOL_10V:
			// 0-10V
			value = (((float) (data&0x3ffff)) * 10.0F / 4.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_18BIT_BIPOL_10V:
		case DAPI_ADDA_MODE_18BIT_BIPOL_10V_TESTMODE:
			// +-10V
			value = (((float) ((data&0x3ffff)) * 10.0F / 4.0F / 65536.0F) - 5.0F ) * 2.0F ;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;
			// 18 Bit !!!!!
			// 18 Bit !!!!!
			// 18 Bit !!!!!


		case DAPI_ADDA_MODE_16BIT_BIPOL_5V:
			// +-5V
			value = (((float) (data&0xffff)) * 10.0F / 65536.0F) - 5.0F;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_0_50mA:
			// 0-5V entspricht 0-50mA (100 Ohm)
			value = (((float) (data&0xffff)) * 50.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_MA;
			break;

		case DAPI_ADDA_MODE_16BIT_0_24mA:
		case DAPI_ADDA_MODE_16BIT_0_20mA:
		case DAPI_ADDA_MODE_16BIT_0_20mA_TESTMODE:
		case DAPI_ADDA_MODE_16BIT_4_20mA:
			if (((DapiModuleParams->hw_interface1 & DAPI_HW_INTERFACE_BIT_SUPPORTED_BY_FIRMWARE) != 0) && \
				((DapiModuleParams->hw_interface1 & DAPI_HW_INTERFACE_PRODUCT_ID_NET) != 0) )
			{
				// NET Serie
				// 0-5V entspricht 0-50mA (100 Ohm) ohne Spannungsverdopplung !
				value = (((float) (data&0xffff)) *50.0F / 65536.0F);
				unit = DAPI_ADDA_UNIT_MA;
			}
			else
			{
				// KEINE NET Serie
				// 0-5V entspricht 0-25mA (100 Ohm) und Spannungsverdopplung !
				value = (((float) (data&0xffff)) *25.0F / 65536.0F);
				unit = DAPI_ADDA_UNIT_MA;
			}
			break;

		default:
			sprintf(debug_msg, "DapiADGetValueModeUnit -> Illegal A/D Mode detected (mode=0x%lx)", (unsigned long) mode);
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
			unit = DAPI_ADDA_UNIT_ILLEGAL;
			break;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiADGetValueModeUnit");
		sprintf(debug_msg, "DapiADGetValueModeUnit(handle32=0x%lx, ch=%ld,  mode=%ld) * ret=%f", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode, value);
		debug_print(debug_msg);
	#endif
	// -----

	if((unsigned long) ret_mode != 0)
	{
		*ret_mode = mode;
	}

	if((unsigned long) ret_unit != 0)
	{
		*ret_unit = unit;
	}

	return value;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDASetMode(ULONG handle32, ULONG ch, ULONG mode)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG mode_rb;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDASetMode(handle32=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDASetMode");
		sprintf(debug_msg, "DapiDASetMode(handle32=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDASetMode -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= ch)		
	{
		sprintf(debug_msg, "DapiDASetMode -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ao);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteByteXXX(handle32, 0x2000 + ch*8 + 2, mode&255);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DA;
		adr |= DEV_A23_16_GROUP_DA_SUB_DA_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		DapiWriteLongXXX(handle32, adr, mode&255);
	}

/*	if((mode & DAPI_ADDA_MODE_DA_DISABLE) == DAPI_ADDA_MODE_DA_DISABLE)
	{
		DapiWriteByteXXX(handle32, 0x2000 + ch*8 + 3, 1);	// Disable D/A Channel
		sprintf(debug_msg,"DapiDASetMode(ch=%ld, mode=0x%lx - disable D/A", ch, mode);
		debug_print(debug_msg);
	}
	if((mode & DAPI_ADDA_MODE_DA_ENABLE) == DAPI_ADDA_MODE_DA_ENABLE)
	{
		DapiWriteByteXXX(handle32, 0x2000 + ch*8 + 3, 0);	// Enable D/A Channel
		sprintf(debug_msg,"DapiDASetMode(ch=%ld, mode=0x%lx - disable D/A", ch, mode);
		debug_print(debug_msg);
	}
*/
	//if((mode & DAPI_ADDA_MODE_DA_DISABLE) != DAPI_ADDA_MODE_DA_DISABLE)
	{
		// Nur testen, wenn nicht disabled wurde

		// illegal mode (readback mode) ? 
		mode_rb = 0;

		if(DapiHandle->is_Adr_32_mode == FALSE)
		{
			mode_rb = DapiReadByteXXX(handle32, 0x2000 + ch*8 + 2);
		}
		else
		{
			adr =  DEV_A31_24_GROUP_DA;
			adr |= DEV_A23_16_GROUP_DA_SUB_DA_RW_MODE;
			adr |= ch << 3;
			adr |= DEV_A07_00_LENGTH_BYTE;
			mode_rb = DapiReadLongXXX(handle32, adr);
		}
		
		if ((mode_rb != (mode&0x7fff)) && ((mode & DAPI_ADDA_MODE_PREVENT_DAPI_MODE_ERROR) == 0))	// Mit 0x8000 kann man die Fehlerausgabe unterdrcken 
		{
			sprintf(debug_msg, "DapiDASetMode -> Setting Mode = 0x%lx  readback from Module=0x%lx", (unsigned long) mode, (unsigned long) mode_rb);
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDAGetMode(ULONG handle32, ULONG ch)
{
	char dapi_caller_text[200];
	unsigned long ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDAGetMode(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ret = DapiDAGetModeXXX(handle32, ch);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiDAGetModeXXX(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	ULONG adr;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDAGetMode-> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			return 0;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= ch)		
	{
		sprintf(debug_msg, "DapiDAGetMode-> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ao);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadByteXXX(handle32, 0x2000 + ch*8 + 2);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DA;
		adr |= DEV_A23_16_GROUP_DA_SUB_DA_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDAGetMode");
		sprintf(debug_msg, "DapiDAGetMode(handle32=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDASet(ULONG handle32, ULONG ch, ULONG data)
{
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDASet(handle32=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	DapiDASetXXX(handle32, ch, data);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDASetXXX(ULONG handle32, ULONG ch, ULONG data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG adr;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDASet");
		sprintf(debug_msg, "DapiDASet(handle32=0x%lx, ch=%ld, data=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDASet-> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= ch)		
	{
		sprintf(debug_msg, "DapiDASet-> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ao);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteWordXXX(handle32, 0x2000 + ch*8, data);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DA;
		adr |= DEV_A23_16_GROUP_DA_SUB_DA_RW_DATA;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;
		DapiWriteLongXXX(handle32, adr, data);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDASetVolt(ULONG handle32, ULONG ch, float volt)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG mode;
	ULONG value;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDASetVolt(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, volt);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDASetVolt");
		sprintf(debug_msg, "DapiDASetVolt(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, volt);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDASetVolt -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= ch)		
	{
		sprintf(debug_msg, "DapiDASetVolt -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ao);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	mode = 	DapiDAGetModeXXX(handle32, ch);

	switch(mode)
	{
		case DAPI_ADDA_MODE_16BIT_UNIPOL_10V:
			value = (ULONG) (3276.8F *2.0F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;

		case DAPI_ADDA_MODE_16BIT_UNIPOL_5V:
			value = (ULONG) (3276.8F * 4.0F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;
		
		case DAPI_ADDA_MODE_16BIT_UNIPOL_2V5:
			value = (ULONG) (3276.8F * 8.0F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;
		
		case DAPI_ADDA_MODE_16BIT_BIPOL_10V:
			value = (ULONG) (32768.0F + 3276.8F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;
		
		case DAPI_ADDA_MODE_16BIT_BIPOL_5V:
			value = (ULONG) (32768.0F + 3276.8F * 2.0F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;

		case DAPI_ADDA_MODE_16BIT_BIPOL_2V5:
			value = (ULONG) (32768.0F + 3276.8F * 4.0F * volt);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;

		default:
			sprintf(debug_msg, "DapiDASetVolt -> Illegal D/A Mode detected (chan=%ld, mode=0x%lx)", (unsigned long) ch, (unsigned long) mode);
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
			break;
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiDASetmA(ULONG handle32, ULONG ch, float data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG mode;
	ULONG value;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDASetmA(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDASetmA");
		sprintf(debug_msg, "DapiDASetmA(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDASetmA -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= ch)		
	{
		sprintf(debug_msg, "DapiDASetmA -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ao);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}
	
	mode = 	DapiDAGetModeXXX(handle32, ch);

	switch(mode)
	{
		case DAPI_ADDA_MODE_16BIT_0_20mA:
			if(data<0) data=0;
			value = (ULONG) (3276.8F * data);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;

		case DAPI_ADDA_MODE_16BIT_4_20mA:
			if(data<4) data=4;
			value = (ULONG) (4096.0F * (data-4.0F));
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;
		
		case DAPI_ADDA_MODE_16BIT_0_24mA:
			if(data<0) data=0;
			value = (ULONG) (65536.0F / 24.0F * data);
			if(value > 0xffff) value = 0xffff;
			DapiDASetXXX(handle32, ch, value );
			break;

		default:
			sprintf(debug_msg, "DapiDASetmA -> Illegal D/A Mode detected (chan=%ld, mode=0x%lx)", (unsigned long) ch, (unsigned long) mode);
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
			break;
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiDAGetValueModeUnit(ULONG handle32, ULONG ch, ULONG *ret_mode, ULONG *ret_unit)
{
	char dapi_caller_text[200];
	float ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiDAGetValueModeUnit(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ret = DapiDAGetValueModeUnitXXX(handle32, ch, ret_mode, ret_unit);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiDAGetValueModeUnitXXX(ULONG handle32, ULONG ch, ULONG *ret_mode, ULONG *ret_unit)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG data;
	ULONG mode;
	ULONG unit;
	float value=0;
	ULONG adr;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
		{
			sprintf(debug_msg, "DapiDAGetValueModeUnit -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			return 0;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_ao <= (ch & 0x7fff))		
	{
		sprintf(debug_msg, "DapiDAGetValueModeUnit -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_ai);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		return 0;
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		data = DapiReadLongXXX(handle32, 0x2000 + ch*8);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DA;
		adr |= DEV_A23_16_GROUP_DA_SUB_DA_RD_DATA_AND_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		data = DapiReadLongXXX(handle32, adr);
	}

	mode = (data >> 16) & 0xff;

	switch(mode)
	{
		case DAPI_ADDA_MODE_16BIT_UNIPOL_10V:
			// 0-10V
			value = (((float) (data&0xffff)) * 10.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_UNIPOL_5V:
			// 0-5V
			value = (((float) (data&0xffff)) * 5.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_BIPOL_10V:
			// +-10V
			value = (((float) (data&0xffff)) * 20.0F / 65536.0F) - 10.0F;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_BIPOL_5V:
			// +-5V
			value = (((float) (data&0xffff)) * 10.0F / 65536.0F) - 5.0F;
			unit = DAPI_ADDA_UNIT_VOLT;
			break;

		case DAPI_ADDA_MODE_16BIT_0_24mA:
			value = (((float) (data&0xffff)) * 24.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_MA;
			break;

		case DAPI_ADDA_MODE_16BIT_0_20mA:
			value = (((float) (data&0xffff)) * 20.0F / 65536.0F);
			unit = DAPI_ADDA_UNIT_MA;
			break;

		case DAPI_ADDA_MODE_16BIT_4_20mA:
			value = (((float) (data&0xffff)) * 16.0F / 65536.0F) + 4.0F;
			unit = DAPI_ADDA_UNIT_MA;
			break;

		default:
			sprintf(debug_msg, "DapiDAGetValueModeUnit -> Illegal D/A Mode detected (chan=%ld)(mode=0x%lx)", (unsigned long) ch, (unsigned long) mode);
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);					
			unit = DAPI_ADDA_UNIT_ILLEGAL;
			break;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiDAGetValueModeUnit");
		sprintf(debug_msg, "DapiDAGetValueModeUnit(handle32=0x%lx, ch=%ld,  mode=%ld) * ret=%f", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode, value);
		debug_print(debug_msg);
	#endif
	// -----

	if((unsigned long) ret_mode != 0)
	{
		*ret_mode = mode;
	}

	if((unsigned long) ret_unit != 0)
	{
		*ret_unit = unit;
	}

	return value;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiTempGet(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	float f;
	ULONG d;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiTempGet(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_TEMP) == 0)
		{
			sprintf(debug_msg, "DapiTempGet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_temp <= (ch & 0x7fff))		
	{
		sprintf(debug_msg, "DapiTempGet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_temp);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	if((ch&0x8000) == 0)
	{
		// Temperatur soll gelesen werden
		if(DapiHandle->is_Adr_32_mode == FALSE)
		{
			d = DapiReadLongXXX(handle32, 0x4000 + ch*8);

			// DEBUG
			#ifdef DEBUG_OUTPUT
				//sprintf(debug_msg, "DapiTempGet ch=%ld ret=0x%lx", ch, d);
				//debug_print(debug_msg);
			#endif
			// -----

			switch((d>>16) & 0xff)
			{
				case 1:	f = ((float) (d&0x7fff)) / 10;break;				// Faktor 10
				case 2:	f = ((float) (d&0x7fff)) / 100;break;				// Faktor 100
				case 0:	f = -9999;break;									// Faktor Sensor disconnected
				default: f=0;
				
				// DEBUG
				#ifdef DEBUG_OUTPUT
					//sprintf(debug_msg, "DapiTempGet 222222 f=%f", f);
					//debug_print(debug_msg);
				#endif
				// -----
			}

			if(((d>>15)&1) != 0) f=-f;			// Negative Temp

			// DEBUG
			#ifdef DEBUG_OUTPUT
				//sprintf(debug_msg, "DapiTempGet 3333333 f=%f", f);
				//debug_print(debug_msg);
			#endif
			// -----
		}
		else
		{
			adr =  DEV_A31_24_GROUP_TEMP;
			adr |= DEV_A23_16_GROUP_TEMP_SUB_RD_DATA;
			adr |= ch << 3;
			adr |= DEV_A07_00_LENGTH_LONG;
			d = DapiReadLongXXX(handle32, adr);

			f = ((float) (d&0x7fffffff)) / 1000.0F;

			if(((d>>31)&1) != 0) f=-f;			// Negative Temp
		}
	}
	else
	{
		// Der Widerstandswert soll gelesen werden
		if(DapiHandle->is_Adr_32_mode == FALSE)
		{
			d = DapiReadWordXXX(handle32, 0x4000 + ch*8 + 6);				// Widerstandswert lesen

			f = ((float) (d&0x7fff)) / 100;								// Faktor 100
			if(((d>>15)&1) != 0) f=-f;									// Negative
		}
		else
		{
			adr =  DEV_A31_24_GROUP_TEMP;
			adr |= DEV_A23_16_GROUP_TEMP_SUB_RD_RESISTOR;
			adr |= ch << 3;
			adr |= DEV_A07_00_LENGTH_WORD;
			d = DapiReadLongXXX(handle32, adr);

			f = ((float) d) / 1000.0F;
		}
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTempGet");
		sprintf(debug_msg, "DapiTempGet(handle32=0x%lx, ch=%ld) * ret=%f", (unsigned long) handle32, (unsigned long) ch, f);
//		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return f;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiCnt48ModeSet(ULONG handle32, ULONG ch, ULONG mode)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiCnt48ModeSet(handle32=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiCnt48ModeSet");
		sprintf(debug_msg, "DapiCnt48ModeSet(handle32=0x%lx, ch=%ld, mode=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_IN48) == 0)
		{
			sprintf(debug_msg, "DapiCnt48ModeSet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_counterin48 <= ch)		
	{
		sprintf(debug_msg, "DapiCnt48ModeSet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_counterin48);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteWordXXX(handle32, 0x5000 + ch*8 + 6, mode);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_CNT8;
		adr |= DEV_A23_16_GROUP_CNT8_SUB_CNT_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;
		DapiWriteLongXXX(handle32, adr, mode);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiCnt48ModeGet(ULONG handle32, ULONG ch)
{	
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiCnt48ModeGet(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_IN48) == 0)
		{
			sprintf(debug_msg, "DapiCnt48ModeGet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_counterin48 <= ch)		
	{
		sprintf(debug_msg, "DapiCnt48ModeGet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_counterin48);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadWordXXX(handle32, 0x5000 + ch*8 + 6);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_CNT8;
		adr |= DEV_A23_16_GROUP_CNT8_SUB_CNT_RW_MODE;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_WORD;
		ret = DapiReadLongXXX(handle32, adr);
	}

	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiCnt48ModeGet");
		sprintf(debug_msg, "DapiCnt48ModeGet(handle32=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiCnt48CounterGet32(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiCnt48CounterGet32(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_IN48) == 0)
		{
			sprintf(debug_msg, "DapiCnt48CounterGet32 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_counterin48 <= ch)		
	{
		sprintf(debug_msg, "DapiCnt48CounterGet32 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_counterin48);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongXXX(handle32, 0x5000 + ch*8);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_CNT8;
		adr |= DEV_A23_16_GROUP_CNT8_SUB_CNT_RD_32;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONG;
		ret = DapiReadLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiCnt48CounterGet32");
		sprintf(debug_msg, "DapiCnt48CounterGet32(handle32=0x%lx, ch=%ld) * ret=0x%lx", (unsigned long) handle32, (unsigned long) ch, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONGLONG DAPI_FUNCTION_PRE DapiCnt48CounterGet48(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONGLONG ret;
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiCnt48CounterGet48(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_IN48) == 0)
		{
			sprintf(debug_msg, "DapiCnt48CounterGet48 -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_counterin48 <= ch)		
	{
		sprintf(debug_msg, "DapiCnt48CounterGet48 -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_counterin48);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		ret = DapiReadLongLongXXX(handle32, 0x5000 + ch*8);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_CNT8;
		adr |= DEV_A23_16_GROUP_CNT8_SUB_CNT_RD_48;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_LONGLONG;
		ret = DapiReadLongLongXXX(handle32, adr);
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiCnt48CounterGet48");
		#ifdef __linux__
			sprintf(debug_msg, "DapiCnt48CounterGet48(handle32=0x%lx, ch=%ld) * ret=0x%llx", (unsigned long) handle32, (unsigned long) ch, ret);
		#else
			sprintf(debug_msg, "DapiCnt48CounterGet48(handle32=0x%lx, ch=%ld) * ret=0x%I64x", (unsigned long) handle32, (unsigned long) ch, ret);
		#endif
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiPulseGenSet(ULONG handle32, ULONG ch, ULONG mode, ULONG par0, ULONG par1, ULONG par2)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiPulseGenSet(handle32=0x%lx, ch=%ld, mode=0x%lx, par0=0x%lx, par1=0x%lx, par2=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode, (unsigned long) par0, (unsigned long) par1, (unsigned long) par2);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiPulseGenSet");
		sprintf(debug_msg, "DapiPulseGenSet(handle32=0x%lx, ch=%ld, mode=0x%lx, par0=0x%lx, par1=0x%lx, par2=0x%lx)", (unsigned long) handle32, (unsigned long) ch, (unsigned long) mode, (unsigned long) par0, (unsigned long) par1, (unsigned long) par2);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_OUT32) == 0)
		{
			sprintf(debug_msg, "DapiPulseGenSet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_counterout32 <= ch)		
	{
		sprintf(debug_msg, "DapiPulseGenSet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_counterout32);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		/*
		debug_set_function_name("DapiPulseGenSet");
		sprintf(debug_msg, "DapiPulseGenSet Reg A = %lx", mode);
		debug_print(debug_msg);
		sprintf(debug_msg, "DapiPulseGenSet Reg B = %lx", par0);
		debug_print(debug_msg);
		sprintf(debug_msg, "DapiPulseGenSet Reg C = %lx", par1);
		debug_print(debug_msg);
		sprintf(debug_msg, "DapiPulseGenSet Reg D = %lx", par2);
		debug_print(debug_msg);
		*/
	#endif
	// -----
	
	DapiWriteLongXXX(handle32, 0x5800 + ch*16 + 0 , mode);
	DapiWriteLongXXX(handle32, 0x5800 + ch*16 + 4 , par0);
	DapiWriteLongXXX(handle32, 0x5800 + ch*16 + 8 , par1);
	DapiWriteLongXXX(handle32, 0x5800 + ch*16 + 12, par2);

	// DEBUG
	#ifdef DEBUG_OUTPUT
		/*
		ULONG ret;
		debug_set_function_name("DapiPulseGenSet");
		ret = DapiReadLongXXX(handle32, 0x5800 + ch*16 + 0);
		sprintf(debug_msg, "DapiPulseGenSet Reg A (readback) = %lx", ret);
		debug_print(debug_msg);
		ret = DapiReadLongXXX(handle32, 0x5800 + ch*16 + 4);
		sprintf(debug_msg, "DapiPulseGenSet Reg B (readback) = %lx", ret);
		debug_print(debug_msg);
		ret = DapiReadLongXXX(handle32, 0x5800 + ch*16 + 8);
		sprintf(debug_msg, "DapiPulseGenSet Reg C (readback) = %lx", ret);
		debug_print(debug_msg);
		ret = DapiReadLongXXX(handle32, 0x5800 + ch*16 + 12);
		sprintf(debug_msg, "DapiPulseGenSet Reg D (readback) = %lx", ret);
		debug_print(debug_msg);
		*/
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiPWMOutSet(ULONG handle32, ULONG ch, float data)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiPWMOutSet(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, data);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiPWMOutSet");
		sprintf(debug_msg, "DapiPWMOutSet(handle32=0x%lx, ch=%ld, data=%f)", (unsigned long) handle32, (unsigned long) ch, data);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_PWM_OUT) == 0)
		{
			sprintf(debug_msg, "DapiPWMOutSet -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_output_pwm <= ch)		
	{
		sprintf(debug_msg, "DapiPWMOutSet -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_output_pwm);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPI_return;
	}

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		DapiWriteByteXXX(handle32, 0x0800 + ch  , (ULONG) data);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_PWM_RW_DATA;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		DapiWriteLongXXX(handle32, adr, (ULONG) data);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 float DAPI_FUNCTION_PRE DapiPWMOutReadback(ULONG handle32, ULONG ch)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	ULONG adr;
	ULONG data;
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiPWMOutReadback(handle32=0x%lx, ch=%ld", (unsigned long) handle32, (unsigned long) ch);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiPWMOutReadback");
		sprintf(debug_msg, "DapiPWMOutReadback(handle32=0x%lx, ch=%ld)", (unsigned long) handle32, (unsigned long) ch);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_PWM_OUT) == 0)
		{
			sprintf(debug_msg, "DapiPWMOutReadback -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal channel ?
	if (DapiModuleParams->anz_output_pwm <= ch)		
	{
		sprintf(debug_msg, "DapiPWMOutReadback -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) ch, (unsigned long) DapiModuleParams->anz_output_pwm);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}

	data = 0;

	if(DapiHandle->is_Adr_32_mode == FALSE)
	{
		data = DapiReadByteXXX(handle32, 0x0800 + ch);
	}
	else
	{
		adr =  DEV_A31_24_GROUP_DO;
		adr |= DEV_A23_16_GROUP_DO_SUB_PWM_RW_DATA;
		adr |= ch << 3;
		adr |= DEV_A07_00_LENGTH_BYTE;
		data = DapiReadLongXXX(handle32, adr);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return (float) data;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiSpecialCommand(ULONG handle32, ULONG cmd, ULONG par1, ULONG par2, ULONG par3)
{
	char dapi_caller_text[200];
	unsigned long ret;

	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiSpecialCommand(handle32=0x%lx, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx)", (unsigned long) handle32, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ret = DapiSpecialCommandXXX(handle32, cmd, par1, par2, par3);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiSpecialCommandXXX(ULONG handle32, ULONG cmd, ULONG par1, ULONG par2, ULONG par3)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	ULONG ret=0;
	ULONG i;
	ULONG bank;
	unsigned char *buff;
	ULONG pos;
	ULONG result;
	ULONG cnt;
	ULONGLONG d64;
	unsigned long adr;
	unsigned char ch;
	
	unsigned long coData = 0;
	unsigned long long coHandle = 0;
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiSpecialCommand");
		sprintf(debug_msg, "DapiSpecialCommand(handle32=0x%lx, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx)", (unsigned long) handle32, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3);
		debug_print(debug_msg);
	#endif
	// -----

	ret = DAPI_ERR_NONE;

	switch(cmd)
	{
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_PWM:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_PWM_OUT) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_PWM....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			// illegal channel ?
			if (DapiModuleParams->anz_output_pwm <= (par2))		
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_PWM....) -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) par2, (unsigned long) DapiModuleParams->anz_output_pwm);
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
			
			// par1 = submcd
			// par2 = channelarea
			// par3 = value
			
			// register addr
			// pwm geht von 0x0800 bis 0x087F	=> 128 channels
			
			// frequenzen dann ab 0x0880 mit 16bit blcken...
			// byte[0] = freq
			// byte[1] = cmd (ee write/read/default) fr diesen block	=> cmds selbe wie DA
			
			
			switch (par1)
			{
				// ------------------------------------
				case DAPI_SPECIAL_PWM_FREQ_LOAD_DFAULT:
					DapiWriteByteXXX(handle32, 0x0880 + par2 + 1, 0x12);		// Auslieferungszustand laden

					#ifndef linux
						dt_wait_us(10*1000);
					#endif

					cnt=100;
					do
					{
						--cnt;
					} while((DapiReadByteXXX(handle32, 0x0880 + par2 + 1) != 0xff) && (cnt!=0));
					
					break;
				
				// ------------------------------------
				case DAPI_SPECIAL_PWM_FREQ_SAVE_EEPROM:
					DapiWriteByteXXX(handle32, 0x0880 + par2 + 1, 0x10);		// Ins EEPROM schreiben
					if(DapiGetLastError() == DAPI_ERR_NONE)
					{
						#ifndef linux
							dt_wait_us(10*1000);
						#endif

						cnt=100;
						do
						{
							--cnt;
							ret = DapiReadByteXXX(handle32, 0x0880 + par2 + 1);

							if(DapiGetLastError() == DAPI_ERR_DEV_IO_RD_WR_ERROR)
							{
								// Bei einem Error wird der gelesene Wert ignoriert....
								// Beim Flash schreiben kann das Modul zeitweise nicht reagieren (liegt am FLASH-ERASE Renesas)
								ret = 0;
								DapiClearLastError();
							}
						} while((ret != 0xff) && (cnt!=0));
					}
					
					break;
					
				// ------------------------------------
				case DAPI_SPECIAL_PWM_FREQ_LOAD_EEPROM:
					DapiWriteByteXXX(handle32, 0x0880 + par2 + 1, 0x11);		// Aus EEPROM laden

					#ifndef linux
						dt_wait_us(10*1000);
					#endif

					cnt=100;
					do
					{
						--cnt;
					} while((DapiReadByteXXX(handle32, 0x0880 + par2 + 1) != 0xff) && (cnt!=0));
				
					break;
					
				// ------------------------------------
				case DAPI_SPECIAL_PWM_FREQ_SET:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "Kanal=%ld - PWM Frequency set= %ld", (unsigned long) (par2&0xff), (unsigned long) (par3&0xff));
					debug_print(debug_msg);

					if(DapiHandle->is_Adr_32_mode == FALSE)
					{
						// Modul Nummer korrekt bestimmen - TODO !!!
						DapiWriteByteXXX(handle32, 0xfe00, (par2/16)&0xff);
						DapiWriteWordXXX(handle32, 0xfe02, ((RO_CPU_CMD_OPTION_PWM_FREQ_SET & 0xff) | ((par3&0xff) << 8)));
					}
					else
					{
						adr =  DEV_A31_24_GROUP_DO;
						adr |= DEV_A23_16_GROUP_DO_SUB_PWM_RW_FREQ;
						adr |= par2 << 3;	// Kanal
						adr |= DEV_A07_00_LENGTH_BYTE;
						DapiWriteLongXXX(handle32, adr, par3&0xff);
					}
					
					break;
				// ------------------------------------
				case DAPI_SPECIAL_PWM_FREQ_READBACK:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "Start Channel=%ld - PWM Frequency readback", (unsigned long) (par2&0xff));
					debug_print(debug_msg);

					if(DapiHandle->is_Adr_32_mode == FALSE)
					{
						DapiWriteByteXXX(handle32, 0xfe00, (par2/16)&0xff);
						DapiWriteByteXXX(handle32, 0xfe02, RO_CPU_CMD_OPTION_PWM_FREQ_READBACK);
						ret = DapiReadByteXXX(handle32, 0xfe05);
					}
					else
					{
						adr =  DEV_A31_24_GROUP_DO;
						adr |= DEV_A23_16_GROUP_DO_SUB_PWM_RW_FREQ;
						adr |= par2 << 3;	// Kanal
						adr |= DEV_A07_00_LENGTH_BYTE;
						ret = DapiReadLongXXX(handle32, adr);
					}
					
					sprintf(debug_msg, "PWM Frequency readback = %d", ret);
					debug_print(debug_msg);
					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_PWM, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------

				break;
			}
			
			break;
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_DA:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DA) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_DA....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			// illegal channel ?
			if (DapiModuleParams->anz_ao <= par2)		
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_DA....) -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) par2, (unsigned long) DapiModuleParams->anz_ao);
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
		
			switch(par1)
			{
				// ------------------------------------
				case DAPI_SPECIAL_DA_PAR_DA_LOAD_DEFAULT:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_LOAD_DEFAULT - Start *** ch=%ld", (unsigned long) par2);
					debug_print(debug_msg);

					DapiWriteByteXXX(handle32, 0x2000 + par2*8 + 7, 0x12);		// Auslieferungszustand laden

					// Dummy values .... for DebugView
					cnt=200;
					ret=0xfe;

					#ifndef linux
						dt_wait_us(10*1000);
					#endif

					cnt=100;
					do
					{
						--cnt;
					} while((DapiReadByteXXX(handle32, 0x2000 + par2*8 + 7) != 0xff) && (cnt!=0));

					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_LOAD_DEFAULT - Ende *** cnt=%ld - ret=0x%lx", (unsigned long) cnt, (unsigned long) ret);
					debug_print(debug_msg);

					break;
				// ------------------------------------
				case DAPI_SPECIAL_DA_PAR_DA_SAVE_EEPROM_CONFIG:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_SAVE_EEPROM_CONFIG - Start *** ch=%ld", (unsigned long) par2);
					debug_print(debug_msg);

					DapiWriteByteXXX(handle32, 0x2000 + par2*8 + 7, 0x10);		// Ins EEPROM schreiben

					// Dummy values .... for DebugView
					cnt=200;
					ret=0xfe;

					if(DapiGetLastError() == DAPI_ERR_NONE)
					{
						#ifndef linux
							dt_wait_us(10*1000);
						#endif

						cnt=100;
						do
						{
							--cnt;
							ret = DapiReadByteXXX(handle32, 0x2000 + par2*8 + 7);

							if(cnt != 0)
							{
								// Der letzte Fehler soll aber dann "stehen" bleiben
								if(DapiGetLastError() == DAPI_ERR_DEV_IO_RD_WR_ERROR)
								{
									// Bei einem Error wird der gelesene Wert ignoriert....
									// Beim Flash schreiben kann das Modul zeitweise nicht reagieren (liegt am FLASH-ERASE Renesas)
									ret = 0;
									DapiClearLastError();
								}
							}
						} while((ret != 0xff) && (cnt!=0));
					}

					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_SAVE_EEPROM_CONFIG - Ende *** cnt=%ld - ret=0x%lx", (unsigned long) cnt, (unsigned long) ret);
					debug_print(debug_msg);

					break;
				// ------------------------------------
				case DAPI_SPECIAL_DA_PAR_DA_LOAD_EEPROM_CONFIG:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_LOAD_EEPROM_CONFIG - Start *** ch=%ld", (unsigned long) par2);
					debug_print(debug_msg);

					DapiWriteByteXXX(handle32, 0x2000 + par2*8 + 7, 0x11);		// Aus EEPROM laden

					// Dummy values .... for DebugView
					cnt=200;
					ret=0xfe;

					#ifndef linux
						dt_wait_us(10*1000);
					#endif

					cnt=100;
					do
					{
						--cnt;
					} while((DapiReadByteXXX(handle32, 0x2000 + par2*8 + 7) != 0xff) && (cnt!=0));

					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "DAPI_SPECIAL_DA_PAR_DA_LOAD_EEPROM_CONFIG - Ende *** cnt=%ld - ret=0x%lx", (unsigned long) cnt, (unsigned long) ret);
					debug_print(debug_msg);

					break;
				// ------------------------------------	
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_DA_PAR_DA, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
			}
			break;

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_AD:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_AD) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			switch(par1)
			{
				// ------------------------------------				
				case DAPI_SPECIAL_AD_READ_MULTIPLE_AD:
			
					// illegal start or end channel ?
					if ((DapiModuleParams->anz_ai <= par2) || (DapiModuleParams->anz_ai <= par3) || (par2>par3))	
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> read multiple byte / Error accessing channel %ld - Module only supports %ld channels", (unsigned long) par3, (unsigned long) DapiModuleParams->anz_ai);
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
					
					ret=DapiReadMultipleBytesXXX(handle32, 0x1000 + par2*4, (par3-par2+1)*4, 1, ((unsigned char*) &global_ad_buffer) + par2*4, (par3-par2+1)*4);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_DEACTIVATE:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
				
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_enable_disable | (0<<8));
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_ACTIVATE:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
				
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_enable_disable | (1<<8));
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_INIT:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
					
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_init | (0<<8));
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_GET_STATUS:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
					
					DapiWriteByteXXX(handle32, 0xfe00, 0);
					DapiWriteByteXXX(handle32, 0xfe02, RO_CPU_CMD_FIFO_GET_STATUS);
					ret = DapiReadByteXXX(handle32, 0xfe05);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_SET_INTERVAL_MS:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}
				
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_set_intervalL | ((par3&255)           <<8));
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_set_intervalH | ((par3>>8) & 255)     <<8);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FIFO_SET_CHANNEL:
					if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
						 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
					{
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD....) -> Software Feature 'AD-FIFO' is not supported");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
					}

					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_enable_chL | ((par3&255)          <<8));
					DapiWriteWordXXX(handle32, 0x1400 + par2*4, RO_FIFO_CMD_fifo_enable_chH | ((par3>>8) & 255)    <<8);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FILTER_SET:
					debug_set_function_name("DapiSpecialCommand");
					sprintf(debug_msg, "Modul=%ld - Filter-level = (par2-Bit8-15) = %ld", (unsigned long) par2&0xff, (unsigned long) (par2>>8)&0xff);
					debug_print(debug_msg);
					if(DapiHandle->is_Adr_32_mode == FALSE)
					{
						DapiWriteByteXXX(handle32, 0xfe00, par2&0xff);
						DapiWriteWordXXX(handle32, 0xfe02, (RO_CPU_CMD_OPTION_FILTER & 0xff) | (((par2>>8)&0xff) << 8));
					}
					else
					{
						adr =  DEV_A31_24_GROUP_AD;
						adr |= DEV_A23_16_GROUP_AD_SUB_AD_RW_FILTER;
						adr |= par2&0xff;	// Modul Nr
						adr |= DEV_A07_00_LENGTH_BYTE;
						DapiWriteLongXXX(handle32, adr, (par2>>8)&0xff);
					}
					

					break;
				// ------------------------------------
				case DAPI_SPECIAL_AD_FILTER_GET:
					debug_set_function_name("DapiSpecialCommand");
					if(DapiHandle->is_Adr_32_mode == FALSE)
					{
					}
					else
					{
						adr =  DEV_A31_24_GROUP_AD;
						adr |= DEV_A23_16_GROUP_AD_SUB_AD_RW_FILTER;
						adr |= par2&0xff;	// Modul Nr
						adr |= DEV_A07_00_LENGTH_BYTE;
						ret = DapiReadLongXXX(handle32, adr);
					}
					sprintf(debug_msg, "Modul=%ld - Filter-level get =  %ld", (unsigned long) par2&0xff, (unsigned long) ret);
					debug_print(debug_msg);
					

					break;
				// ------------------------------------	
				default: 
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_AD, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__); 
					break;
				// ------------------------------------
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_SET_DIR_DX_8:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DX) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_8....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			// illegal channel ?
			if (DapiModuleParams->anz_dx <= par1)
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_8....) -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) par1, (unsigned long) DapiModuleParams->anz_dx);
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
			
			// illegal channel parameter ? 
			if ((par1 & 0x3f) != 0)
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_8....) -> Illegal channel parameter for this command (Valid: 0, 8, 16, ..)");
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
			
			DapiWriteLongXXX(handle32, 0x100 + par1/64, par2);
			break;

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_SET_DIR_DX_1:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type OR
				// module does not support this command
				if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DX) == 0) ||\
					 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_DX_1_MODE) == 0) )
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_1....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			// illegal channel ?
			if (DapiModuleParams->anz_dx <= par1)
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_1....) -> Error accessing channel %ld - Module only supports %ld channels", (unsigned long) par1, (unsigned long) DapiModuleParams->anz_dx);
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
			
			// illegal channel parameter ? 
			if ((par1 & 0x07) != 0)	
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SET_DIR_DX_1....) -> Illegal channel parameter for this command (Valid: 0, 8, 16, ..)");
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
			
			DapiWriteLongXXX(handle32, 0x120 + par1/8, par2);
			break;

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
			
		case DAPI_SPECIAL_CMD_GET_MODULE_CONFIG:
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)
			{
				switch(par1)
				{
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DO:				ret = DapiModuleParams->anz_do;			break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI:				ret = DapiModuleParams->anz_di;			break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI_FF:			ret = DapiModuleParams->anz_di_ff;		break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI_COUNTER:		ret = DapiModuleParams->anz_di_counter;	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DX:				ret = DapiModuleParams->anz_dx;			break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_AD:				ret = DapiModuleParams->anz_ai;			break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DA:				ret = DapiModuleParams->anz_ao;			break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_STEPPER:		ret = DapiModuleParams->anz_stepper;	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_TEMP:			ret = DapiModuleParams->anz_temp;		break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_CNT48:			ret = DapiModuleParams->anz_counterin48;	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_PULSE_GEN:		ret = DapiModuleParams->anz_counterout32;	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_PWM_OUT:		ret = DapiModuleParams->anz_output_pwm; break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_SW_FEATURE1:	ret = DapiModuleParams->sw_feature1; 	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_HW_INTERFACE1:	ret = DapiModuleParams->hw_interface1; 	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_SW_CLASS:		ret = DapiModuleParams->sw_class; 		break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_HW_GROUP:		ret = DapiModuleParams->hw_group; 		break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_MODULE_ID:		ret = DapiModuleParams->moduleID; 		break;
					default:
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG....) -> Module does not support this command");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
				}
			}
			else
			{
				switch(par1)
				{
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DO:				ret = DapiReadByteXXX(handle32, 0xff00); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI:				ret = DapiReadByteXXX(handle32, 0xff02); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI_FF:			ret = DapiReadByteXXX(handle32, 0xff0c); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI_COUNTER:		ret = DapiReadByteXXX(handle32, 0xff0e); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DX:				ret = DapiReadByteXXX(handle32, 0xff04); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_AD:				ret = DapiReadByteXXX(handle32, 0xff08); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DA:				ret = DapiReadByteXXX(handle32, 0xff06); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_STEPPER:		ret = DapiReadByteXXX(handle32, 0xff0a); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_TEMP:			ret = DapiReadByteXXX(handle32, 0xff10); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_CNT48:			ret = DapiReadByteXXX(handle32, 0xff12); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_PULSE_GEN:		ret = DapiReadByteXXX(handle32, 0xff14); break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_PWM_OUT:		ret = DapiReadByteXXX(handle32, 0xff16); break;

					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_SW_CLASS:		ret = DapiReadLongXXX(handle32, 0xfff0);	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_HW_GROUP:		ret = DapiReadLongXXX(handle32, 0xffec);	break;

					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_SW_FEATURE1:	ret = 0;	break;
					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_HW_INTERFACE1:	ret = 0;	break;

					case DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_MODULE_ID:		ret = DapiHandle->moduleID;break;//DapiModuleParams->moduleID; 		break;

					// Ab hier Error ausgeben - Kommando gibt es nicht !
					default:
						sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG....) -> Module does not support this command");
						DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
						return 0;
						break;
				}
			} // if
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_GET_MODULE_VERSION:
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)
			{
				switch(par1)
				{
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_0:		ret = (ULONG) DapiModuleParams->fw_version[0] - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_1:		ret = (ULONG) DapiModuleParams->fw_version[1] - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_2:		ret = (ULONG) DapiModuleParams->fw_version[2] - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_3:		ret = (ULONG) DapiModuleParams->fw_version[3] - '0';break;
				}
			}
			else
			{
				switch(par1)
				{
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_0:		ret = DapiReadByteXXX(handle32, 0xfff4) - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_1:		ret = DapiReadByteXXX(handle32, 0xfff5) - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_2:		ret = DapiReadByteXXX(handle32, 0xfff6) - '0';break;
					case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_3:		ret = DapiReadByteXXX(handle32, 0xfff7) - '0';break;
				}
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
/*		case DAPI_SPECIAL_CMD_GET_MODULE_XXXXXVERSION:
			switch(par1)
			{
				case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_0:		ret = (ULONG) DapiModuleParams->fw_version[0] - '0';break;
				case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_1:		ret = (ULONG) DapiModuleParams->fw_version[1] - '0';break;
				case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_2:		ret = (ULONG) DapiModuleParams->fw_version[2] - '0';break;
				case DAPI_SPECIAL_GET_MODULE_PAR_VERSION_3:		ret = (ULONG) DapiModuleParams->fw_version[3] - '0';break;
			}
			break;
*/			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_MODULE_REBOOT:
			DapiWriteByteXXX(handle32, 0xfe11, 16);      				// Restart Module
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_MODULE_RESCAN:
			if(DapiHandle->moduleID != USB_CANOPEN_STICK)
			{
				DapiWriteByteXXX(handle32, 0xfe11, 78);      				// ReSCAN Module
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_RESTART_CHECK_MODULE_CONFIG:
			restart_check_module_config(handle32);
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------

		case DAPI_SPECIAL_CMD_TIMEOUT:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			switch(par1)
			{
				case DAPI_SPECIAL_TIMEOUT_SET_VALUE_SEC:
					DapiWriteWordXXX(handle32, 0xfd02, par2*10+ (par3%10));
					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_ACTIVATE:
					DapiWriteByteXXX(handle32, 0xfd00,1);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DEACTIVATE:
					DapiWriteByteXXX(handle32, 0xfd00,0);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_GET_STATUS:
					ret = DapiReadByteXXX(handle32, 0xfd01);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DO_VALUE_LOAD_DEFAULT:
					// module supports new Software-Features ?
					if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
					{
						// module does not have the accessed I/O-type
						if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK) == 0)
						{
							sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK....) -> Module does not support this command");
							DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
							return 0;
						}
					}

					DapiWriteByteXXX(handle32, 0xfd04, 1);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DO_VALUE_MASK_WR_CLR32:
					// module supports new Software-Features ?
					if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
					{
						// module does not have the accessed I/O-type
						if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK) == 0)
						{
							sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK....) -> Module does not support this command");
							DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
							return 0;
						}
					}

					DapiWriteLongXXX(handle32, 0x0700 + ((par2 >> 3) & 0x1f), par3);	// 0x0700 - Value to Clear

					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DO_VALUE_MASK_RD_CLR32:
					// module supports new Software-Features ?
					if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
					{
						// module does not have the accessed I/O-type
						if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK) == 0)
						{
							sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK....) -> Module does not support this command");
							DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
							return 0;
						}
					}

					ret = DapiReadLongXXX(handle32, 0x0700 + ((par2 >> 3) & 0x1f));	// 0x0700 - Value to Clear

					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DO_VALUE_MASK_WR_SET32:
					// module supports new Software-Features ?
					if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
					{
						// module does not have the accessed I/O-type
						if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK) == 0)
						{
							sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK....) -> Module does not support this command");
							DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
							return 0;
						}
					}

					DapiWriteLongXXX(handle32, 0x0720 + ((par2 >> 3) & 0x1f), par3);	// 0x0720 - Value to Set

					break;
				// ------------------------------------
				case DAPI_SPECIAL_TIMEOUT_DO_VALUE_MASK_RD_SET32:
					// module supports new Software-Features ?
					if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
					{
						// module does not have the accessed I/O-type
						if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK) == 0)
						{
							sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, DAPI_SW_FEATURE_BIT_AUTO_OUTPUTS_OFF_TIMEOUT_WITH_MASK....) -> Module does not support this command");
							DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
							return 0;
						}
					}

					ret = DapiReadLongXXX(handle32, 0x0720 + ((par2 >> 3) & 0x1f));	// 0x0720 - Value to Set

					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_TIMEOUT, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1);
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_FUNCTION, 0, debug_msg, __FILE__, __LINE__);
					return 0;
					break;
				// ------------------------------------
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_DI:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_DI....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			switch(par1)
			{
				case DAPI_SPECIAL_DI_FF_FILTER_VALUE_SET:
					DapiWriteByteXXX(handle32, 0xfd10, par2);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_DI_FF_FILTER_VALUE_GET:
					ret = DapiReadByteXXX(handle32, 0xfd10);
					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_DI, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_WATCHDOG:		
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_WATCHDOG) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_WATCHDOG....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			switch(par1)
			{
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_SET_TIMEOUT_MSEC:
					DapiWriteLongXXX(handle32, 0xe004, par2);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_GET_TIMEOUT_MSEC:
					ret = DapiReadLongXXX(handle32, 0xe004);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_GET_STATUS:
					ret=DapiReadByteXXX(handle32, 0xe000);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_GET_WD_COUNTER_MSEC:
					ret = DapiReadLongXXX(handle32, 0xe008);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_GET_TIMEOUT_RELAIS_COUNTER_MSEC:
					ret = DapiReadLongXXX(handle32, 0xe00c);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_SET_TIMEOUT_REL1_COUNTER_MSEC:
					DapiWriteLongXXX(handle32, 0xe008, par2);
					break;
				// ------------------------------------
				case DAPI_SPECIAL_WATCHDOG_SET_TIMEOUT_REL2_COUNTER_MSEC:
					DapiWriteLongXXX(handle32, 0xe00c, par2);
					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_WATCHDOG, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_COUNTER:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI_CNT) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_COUNTER....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			// Abfragen sind absichtlich hintereinander. Es kann ja in Zukunft Counter Commandos geben, die nicht fr das LATCH Modul sind ..

			// module does not have the accessed I/O-type
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_DI_CNT_LATCH) == 0)
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_COUNTER_LATCH....) -> Module does not support this command");
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}

			switch(par1)
			{
				case DAPI_SPECIAL_COUNTER_LATCH_ALL:
					DapiWriteByteXXX(handle32, 0xfe12, 0x19);					// Latch all Counter
					break;
				// ------------------------------------
				case DAPI_SPECIAL_COUNTER_LATCH_ALL_WITH_RESET:
					DapiWriteByteXXX(handle32, 0xfe12, 0x1a);					// Latch all Counter WITH RESET
					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_COUNTER, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------
			}
			break;

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_CNT48:
			// module supports new Software-Features ?
			if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
			{
				// module does not have the accessed I/O-type
				if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_CNT_IN48) == 0)
				{
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_CNT48....) -> Module does not support this command");
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
					return 0;
				}
			}
			
			switch(par1)
			{
				// ------------------------------------
				case DAPI_SPECIAL_CNT48_RESET_SINGLE:
					DapiWriteByteXXX(handle32, 0x5000 + par2*8, 0x00);			// Reset Counter Nr. "par2"
					break;
				// ------------------------------------
				case DAPI_SPECIAL_CNT48_RESET_GROUP8:
					DapiWriteByteXXX(handle32, 0x5003 + (par2/8)*0x40, 0x00);	// Reset Counter Group Nr. "par2"
					break;
				// ------------------------------------
				case DAPI_SPECIAL_CNT48_LATCH_GROUP8:
					DapiWriteByteXXX(handle32, 0x5002 + (par2/8)*0x40, 0x00);	// Latch Counter Group Nr. "par2"
					break;
				// ------------------------------------
				case DAPI_SPECIAL_CNT48_DI_GET1:
					ret = DapiReadByteXXX(handle32, 0x5007 + (par2*8)) & 1;		// Bit 0 Latch Counter Group Nr. "par2"

					//sprintf(debug_msg,"DAPI_SPECIAL_CMD_CNT48: Erase State 2   par2=%ld    ret * ret=0x%lx", (unsigned long) par2, (unsigned long) ret);
					//debug_print(debug_msg);

					break;
				// ------------------------------------	
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_CNT48, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1);
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------
			}
			break;
			
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		case DAPI_SPECIAL_CMD_SOFTWARE_FIFO:
			// module does NOT support new Software-Features ? OR
			// module does NOT support new Software-FIFO command ?
			if ( ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) == 0) || \
				 ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SOFTWARE_FIFO) == 0) )
			{
				sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SOFTWARE_FIFO....) -> Software Feature 'AD-FIFO' is not supported");
				DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_SW_FEATURE, 0, debug_msg, __FILE__, __LINE__);
				return 0;
			}
		
			switch(par1)
			{
				// ------------------------------------
				case DAPI_SPECIAL_SOFTWARE_FIFO_ACTIVATE:
					DapiWriteByteXXX(handle32, 0xfe30, 0x01);					// activate software fifo
					break;
				// ------------------------------------
				case DAPI_SPECIAL_SOFTWARE_FIFO_DEACTIVATE:
					DapiWriteByteXXX(handle32, 0xfe30, 0x00);					// deactivate software fifo
					break;	
				// ------------------------------------
				case DAPI_SPECIAL_SOFTWARE_FIFO_GET_STATUS:
					ret = DapiReadByteXXX(handle32, 0xfe30) & 1;					// read software-fifo status (activated / deactivated)
					break;
				// ------------------------------------
				default:
					sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_SPECIAL_CMD_SOFTWARE_FIFO, %d (dez)....) -> par1 = %d (dez) is not supported", par1, par1); 
					DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
					break;
				// ------------------------------------
			}
			break;

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------

		
		#ifndef __linux__
			#include "delib_Dapi_di_do_ad_da_befehle_ext.cpp"
		#endif 
		// Nicht fr Linux
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiSpecialCommand");
		sprintf(debug_msg, "return=0x%lx", (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifndef __linux__
DAPI_FUNCTION_PRE64 	ULONG DAPI_FUNCTION_PRE DapiScanModule(ULONG moduleID, ULONG cmd)
	{
		ULONG ret=0;
		ULONG BusType;
		ULONG i;
		ULONG handle32;
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiScanModule(0x%x, 0x%x)", moduleID, cmd);
	DapiSetErrorCallerParamsSet(0, dapi_caller_text, sizeof(dapi_caller_text));


		if(cmd==DAPI_SCANMODULE_GET_MODULES_AVAILABLE)
		{

			BusType = DapiInternGetBusType(moduleID);
			if(( BusType == bus_FTDI_DUAL) ||
				(BusType == bus_SeriellFTDI) ||
				(BusType == bus_FTDI_SINGLE_SER) ||
				(BusType == bus_FTDI_SINGLE_PAR))
			{
	#ifdef USE_FTDI
				ret = DapiScanFTDIModulesGetModulesAvailable(moduleID);
	#endif
			}
			else
			{
				for(i=0;i!=8;++i)
				{
					handle32 = DapiOpenModuleXXX(moduleID, i);
					if(handle32 != 0)
					{
						ret |= (1<<i);
						DapiCloseModuleXXX(handle32);
					}
				}
			}
		}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
	}
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiStepperCommand(ULONG handle32, ULONG motor, ULONG cmd, ULONG par1, ULONG par2, ULONG par3, ULONG par4)
{
	ULONG ret;
	ULONG adr_base = 0x3000 | motor*0x40;
	ULONG pos;
	unsigned char buff[20];
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiStepperCommand(handle32=0x%lx, motor=%ld, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx, par4=0x%lx)", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3, (unsigned long) par4);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	ret=0;
	pos = 0;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_STEPPER) == 0)
		{
			sprintf(debug_msg, "DapiStepperCommand -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal motor ?
	if (DapiModuleParams->anz_stepper <= motor)		
	{
		sprintf(debug_msg, "DapiStepperCommand -> Error accessing motor %ld - Module only supports %ld motors", (unsigned long) motor, (unsigned long) DapiModuleParams->anz_stepper);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	buff[pos++] = (UCHAR) ((par3)     & 0xff);
	buff[pos++] = (UCHAR) ((par3>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par3>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par3>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par2)     & 0xff);
	buff[pos++] = (UCHAR) ((par2>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par2>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par2>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par1)     & 0xff);
	buff[pos++] = (UCHAR) ((par1>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par1>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par1>>24) & 0xff);

	buff[pos++] = (UCHAR) ((cmd)     & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>16) & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>24) & 0xff);

	DapiWriteMultipleBytesXXX(handle32, adr_base + 0x10 , pos, 1, buff , pos);

	switch (cmd & 0xc0000000)
	{
		// ------------------------------------
		case DAPI_STEPPER_RETURN_0_BYTES:
			ret=0;
			break;
		// ------------------------------------
		case DAPI_STEPPER_RETURN_1_BYTES:
			ret=DapiReadByteXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------
		case DAPI_STEPPER_RETURN_2_BYTES:
			ret=DapiReadWordXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------
		case DAPI_STEPPER_RETURN_4_BYTES:
			ret=DapiReadLongXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------
		default:
			sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_STEPPER_RETURN...) -> Module does not support this command"); 
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiStepperCommand");
		sprintf(debug_msg, "DapiStepperCommand(handle32=0x%lx, motor=%ld, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx, par4=0x%lx) * ret=0x%lx", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3, (unsigned long) par4, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiStepperCommandEx(ULONG handle32, ULONG motor, ULONG cmd, ULONG par1, ULONG par2, ULONG par3, ULONG par4, ULONG par5, ULONG par6, ULONG par7)
{
	ULONG ret;
	ULONG adr_base = 0x3000 | motor*0x40;
	unsigned char buff[40];
	ULONG pos;
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiStepperCommandEx(handle32=0x%lx, motor=%ld, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx, par4=0x%lx, par5=0x%lx, par6=0x%lx, par7=0x%lx)", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3, (unsigned long) par4, (unsigned long) par5, (unsigned long) par6, (unsigned long) par7);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	
	ret=0;
	pos = 0;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_STEPPER) == 0)
		{
			sprintf(debug_msg, "DapiStepperCommandEx -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal motor ?
	if (DapiModuleParams->anz_stepper <= motor)		
	{
		sprintf(debug_msg, "DapiStepperCommandEx -> Error accessing motor %ld - Module only supports %ld motors", (unsigned long) motor, (unsigned long) DapiModuleParams->anz_stepper);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	buff[pos++] = (UCHAR) ((par7)     & 0xff);
	buff[pos++] = (UCHAR) ((par7>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par7>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par7>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par6)     & 0xff);
	buff[pos++] = (UCHAR) ((par6>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par6>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par6>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par5)     & 0xff);
	buff[pos++] = (UCHAR) ((par5>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par5>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par5>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par4)     & 0xff);
	buff[pos++] = (UCHAR) ((par4>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par4>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par4>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par3)     & 0xff);
	buff[pos++] = (UCHAR) ((par3>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par3>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par3>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par2)     & 0xff);
	buff[pos++] = (UCHAR) ((par2>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par2>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par2>>24) & 0xff);

	buff[pos++] = (UCHAR) ((par1)     & 0xff);
	buff[pos++] = (UCHAR) ((par1>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((par1>>16) & 0xff);
	buff[pos++] = (UCHAR) ((par1>>24) & 0xff);

	buff[pos++] = (UCHAR) ((cmd)     & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>8)  & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>16) & 0xff);
	buff[pos++] = (UCHAR) ((cmd>>24) & 0xff);

	DapiWriteMultipleBytesXXX(handle32, adr_base + 0x00 , pos, 1, buff , pos);

	switch (cmd & 0xc0000000)
	{
		// ------------------------------------
		case DAPI_STEPPER_RETURN_0_BYTES:
			ret=0;
			break;
		// ------------------------------------
		case DAPI_STEPPER_RETURN_1_BYTES:
			ret=DapiReadByteXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------		
		case DAPI_STEPPER_RETURN_2_BYTES:
			ret=DapiReadWordXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------
		case DAPI_STEPPER_RETURN_4_BYTES:
			ret=DapiReadLongXXX(handle32, adr_base | 0x20);;
			break;
		// ------------------------------------
		default:
			sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_STEPPER_RETURN...) -> Module does not support this command"); 
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
			break;
		// ------------------------------------
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiStepperCommandEx");
		sprintf(debug_msg, "DapiStepperCommandEx(handle32=0x%lx, motor=%ld, cmd=0x%lx, par1=0x%lx, par2=0x%lx, par3=0x%lx, par4=0x%lx, par5=0x%lx, par6=0x%lx, par7=0x%lx) * ret=0x%lx", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd, (unsigned long) par1, (unsigned long) par2, (unsigned long) par3, (unsigned long) par4, (unsigned long) par5, (unsigned long) par6, (unsigned long) par7, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiStepperGetStatus(ULONG handle32, ULONG motor, ULONG cmd)
{
	ULONG ret;
	ULONG adr_base = 0x3000 | motor*0x40;
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiStepperGetStatus(handle32=0x%lx, motor=%ld, cmd=0x%lx)", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPIx_return(0);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;
	ret=0;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_STEPPER) == 0)
		{
			sprintf(debug_msg, "DapiStepperGetStatus -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPIx_return(0);
		}
	}
	
	// illegal motor ?
	if (DapiModuleParams->anz_stepper <= motor)		
	{
		sprintf(debug_msg, "DapiStepperGetStatus -> Error accessing motor %ld - Module only supports %ld motors", (unsigned long) motor, (unsigned long) DapiModuleParams->anz_stepper);
		DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_CHANNEL, 0, debug_msg, __FILE__, __LINE__);
		DAPIx_return(0);
	}
	
	switch(cmd)
	{
		// ------------------------------------
		case DAPI_STEPPER_STATUS_GET_POSITION:
			// Get Motorposition
			ret = DapiReadLongXXX(handle32, adr_base | 0x28);
			break;
		// ------------------------------------
		case DAPI_STEPPER_STATUS_GET_SWITCH:
			// Get Endswitches
			ret = DapiReadByteXXX(handle32, adr_base | 0x2f);
			break;
		// ------------------------------------
		case DAPI_STEPPER_STATUS_GET_ACTIVITY:
			// Get Endswitches
			ret = DapiReadByteXXX(handle32, adr_base | 0x2e);
			break;
		// ------------------------------------
		default:
			sprintf(debug_msg, "DapiSpecialCommand(handle32, DAPI_STEPPER_STATUS_GET...) -> Module does not support this command"); 
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_MODE, 0, debug_msg, __FILE__, __LINE__);
			break;
		// ------------------------------------
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiStepperGetStatus");
		sprintf(debug_msg, "DapiStepperGetStatus(handle32=0x%lx, motor=%ld, cmd=0x%lx) * ret=0x%lx", (unsigned long) handle32, (unsigned long) motor, (unsigned long) cmd, (unsigned long) ret);
		debug_print(debug_msg);
	#endif
	// -----

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiWatchdogEnable(ULONG handle32)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiWatchdogEnable(handle32=0x%lx)", (unsigned long) handle32);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));
	
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiWatchdogEnable");
		sprintf(debug_msg, "DapiWatchdogEnable(handle32=0x%lx)", (unsigned long) handle32);
		debug_print(debug_msg);
	#endif
	// -----

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_WATCHDOG) == 0)
		{
			sprintf(debug_msg, "DapiWatchdogEnable -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}

	DapiWriteByteXXX(handle32, 0xe000, 0x23);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiWatchdogDisable(ULONG handle32)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiWatchdogDisable(handle32=0x%lx)", (unsigned long) handle32);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiWatchdogDisable");
		sprintf(debug_msg, "DapiWatchdogDisable(handle32=0x%lx)", (unsigned long) handle32);
		debug_print(debug_msg);
	#endif
	// -----
	
	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_WATCHDOG) == 0)
		{
			sprintf(debug_msg, "DapiWatchdogEnable -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	DapiWriteByteXXX(handle32, 0xe000, 0x12);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 void DAPI_FUNCTION_PRE DapiWatchdogRetrigger(ULONG handle32)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------
	
	sprintf(dapi_caller_text, "DapiWatchdogRetrigger(handle32=0x%lx)", (unsigned long) handle32);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		DAPI_return;
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiWatchdogRetrigger");
		sprintf(debug_msg, "DapiWatchdogRetrigger(handle32=0x%lx)", (unsigned long) handle32);
		debug_print(debug_msg);
	#endif
	// -----
	
	// module supports new Software-Features ?
	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)	
	{
		// module does not have the accessed I/O-type
		if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_CFG_WATCHDOG) == 0)
		{
			sprintf(debug_msg, "DapiWatchdogRetrigger -> Module does not support this command");
			DapiSetError(DAPI_ERR_GEN_NOT_SUPPORTED_IO_TYPE, 0, debug_msg, __FILE__, __LINE__);
			DAPI_return;
		}
	}
	
	DapiWriteByteXXX(handle32, 0xe001, 0x34);

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

DAPI_FUNCTION_PRE64 ULONG DAPI_FUNCTION_PRE DapiGetFirmwareVersion(ULONG handle32, char *buff, ULONG buff_size)
{
	DAPI_HANDLE *DapiHandle;
	DAPI_MODULE_PARAMS *DapiModuleParams;
	char temp[20];
	ULONGLONG handle_ull;
	ULONG ret = 0;
	char debug_msg[200];
	char dapi_caller_text[200];
	
	// --------------------------------
	DapiInternPreventSecondCallStart();
	// --------------------------------

	sprintf(dapi_caller_text, "DapiGetFirmwareVersion(handle32=0x%lx)", (unsigned long) handle32);
	DapiSetErrorCallerParamsSet(handle32, dapi_caller_text, sizeof(dapi_caller_text));

	if (DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);
		DAPIx_return(RETURN_ERROR);
	}

	DapiModuleParams = (DAPI_MODULE_PARAMS *) DAPI_PTR_SIZE DapiHandle->addr_dapi_module_params;

	if ((DapiModuleParams->sw_feature1 & DAPI_SW_FEATURE_BIT_SUPPORTED_BY_FIRMWARE) != 0)
	{
		sprintf(temp, "%c%c.%c%c", DapiModuleParams->fw_version[0], DapiModuleParams->fw_version[1], DapiModuleParams->fw_version[2], DapiModuleParams->fw_version[3]);
		//debug_info("module-params -> temp=%s\n", temp);
	}
	else
	{
		sprintf(temp, "%c%c.%c%c", DapiReadByteXXX(handle32, 0xfff4), DapiReadByteXXX(handle32, 0xfff5), DapiReadByteXXX(handle32, 0xfff6), DapiReadByteXXX(handle32, 0xfff7));
		//debug_info("register -> temp=%s\n", temp);
	}

	if ((ret = strcpy_n(buff, buff_size, temp, (unsigned long) strlen(temp))) != RETURN_OK)
	{
		sprintf(debug_msg, "DapiGetFirmwareVersion -> Error while strcpy - buff too small");
		DapiSetError(DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR, 0, debug_msg, __FILE__, __LINE__);
	}

	// --------------------------------
	DapiInternPreventSecondCallEnd();
	// --------------------------------

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
